from pathlib import Path

import pytest
from timbal import Flow
from timbal.state import RunContext
from timbal.state.savers import JSONLSaver


def test_jsonl_init():
    test_file_dir = Path(__file__).parent
    jsonl_path = test_file_dir / "test_jsonl_init.jsonl"
    JSONLSaver(jsonl_path)

    assert jsonl_path.exists()
    assert jsonl_path.stat().st_size == 0

    # Cleanup
    jsonl_path.unlink()


@pytest.mark.asyncio
async def test_jsonl_put():
    test_file_dir = Path(__file__).parent
    jsonl_path = test_file_dir / "test_jsonl_put.jsonl"
    jsonl_saver = JSONLSaver(jsonl_path)

    with open(jsonl_path) as f:
        jsonl_lines_cnt = len(f.readlines())

    flow = (
        Flow()
        .add_llm(model="gpt-4o-mini", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "response")
        .compile(state_saver=jsonl_saver)
    )

    await flow.complete(prompt="Hello, world!")

    # Assert there is exactly one line in the file
    with open(jsonl_path) as f:
        assert len(f.readlines()) == jsonl_lines_cnt + 1

    # Cleanup
    jsonl_path.unlink()


@pytest.mark.asyncio
async def test_jsonl():
    test_file_dir = Path(__file__).parent
    jsonl_path = test_file_dir / "test_jsonl.jsonl"
    jsonl_saver = JSONLSaver(jsonl_path)

    flow = (
        Flow()
        .add_llm(model="gpt-4o-mini", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "response")
        .compile(state_saver=jsonl_saver)
    )

    flow_output_event = await flow.complete(prompt="My name is David")

    flow_output_event = await flow.complete(
        context=RunContext(parent_id=flow_output_event.run_id),
        prompt="What is my name?"
    )

    flow_output_text = flow_output_event.output["response"].content[0].text
    assert "david" in flow_output_text.lower(), "Response should mention that my name is David"

    # Cleanup
    jsonl_path.unlink()



import os
from typing import Any, Literal

import structlog
from anthropic import AsyncAnthropic
from dotenv import load_dotenv

from ...types import Field, Message, Tool

logger = structlog.get_logger("timbal.steps.llms.anthropic_llm")

load_dotenv()

async def handler(
    # This is not actually used inside the function. It's a hack to have it as an argument
    # so we can use it when mapping data in the flow.
    prompt: Message = Field(default=None, description="Message to send to the LLM."), # noqa: ARG001
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    memory: list[Message] = Field(description="A list containing the conversation history between the user and the LLM"),
    model: Literal["claude-3-5-sonnet-20241022", "claude-3-5-haiku-20241022", "claude-3-opus-20240229", "claude-3-sonnet-20240229", "claude-3-haiku-20240307"] | str = Field(default="claude-3-haiku-20240307", description="Name of the LLM model to use"),
    tools: list[Tool | dict] = Field(default=None, description="List of tools/functions that the LLM can call"),
    tool_choice: dict[str, Any] | str = Field(default = {"type": "auto"}, description="How the model should use the provided tools"),
    max_tokens: int = Field(default=1024, description="The maximum number of tokens in the response"),
    stop: str | list[str] = Field(default=None, description="Where the model will stop generating further tokens"),
    temperature: float = Field(default=1, description="Amount of randomness injected into the response. Ranges from 0.0 to 1.0"),
    top_k: int = Field(default=None, description="Only sample from the top K options for each subsequent token. Must be greater than 0"),
    top_p: float = Field(default=None, description="Nucleus sampling parameter. Ranges from 0.0 to 1.0"),
    json_schema: dict = Field(default=None, description="The JSON input schema that the model must output"),
) -> Any:
    """This handler manages interactions with Anthropic's API, processing requests with 
    appropriate parameters and returning streaming responses.

    Args:
        prompt: Hack to pass an input to the LLM.
        memory: Chat history containing user and LLM messages.
        system_prompt: Instructions for the LLM to follow.
        model: Name of the Anthropic model to use.
        tools: List of tools/functions the LLM can call.
        tool_choice: How the model should use the provided tools
        max_tokens: Maximum number of tokens in the response. Defaults to 1024.
        stop: Sequences where token generation should stop.
        temperature: Sampling temperature between 0 and 1. Defaults to 1.
        top_k: Only sample from the top K options for each subsequent token.
        top_p: Nucleus sampling parameter. Ranges from 0.0 to 1.0.
        json_schema: JSON input schema that the model must output.

    Returns:
        Any: The streaming response from the Anthropic API.
    """
    client = AsyncAnthropic(api_key=os.getenv("ANTHROPIC_API_KEY"))

    # Get default values for parameters
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    model = model.default if hasattr(model, 'default') else model
    tools = tools.default if hasattr(tools, 'default') else tools
    tool_choice = tool_choice.default if hasattr(tool_choice, 'default') else tool_choice
    max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    stop = stop.default if hasattr(stop, 'default') else stop
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    top_k = top_k.default if hasattr(top_k, 'default') else top_k
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    json_schema = json_schema.default if hasattr(json_schema, 'default') else json_schema

    messages = []
    for msg in memory:
        anthropic_msg = msg.to_anthropic_input()
        messages.append(anthropic_msg)

    kwargs = {
        "model": model,
        "max_tokens": max_tokens,
        "messages": messages,
        "stream": True,
        "temperature": temperature,
    }

    # Optional parameters
    if system_prompt:
        kwargs["system"] = system_prompt
    if stop:
        kwargs["stop_sequences"] = stop
    if top_k:
        kwargs["top_k"] = top_k
    if top_p:
        kwargs["top_p"] = top_p

    # Change parameters to Anthropic format
    if isinstance(tool_choice, str):
        tool_choice = {"type": tool_choice}

    if isinstance(stop, list):
        kwargs["stop_sequences"] = stop[0]

    if tools:
        if "haiku" in model:
            logger.warning("This model does not support multiple tool calls. Only one tool can be called at a time.", model=model)
        
        for tool in tools:
            if isinstance(tool, Tool):
                tool = tool.to_anthropic()
        kwargs["tools"] = tools
        kwargs["tool_choice"] = tool_choice

    if json_schema:
        kwargs["tools"] = [
            {
                "name": "print_json",
                "description": "Print in a JSON format",
                "input_schema": {
                    "type": "object",
                    "properties": json_schema
                }
            }
        ]
        kwargs["tool_choice"] = {"type": "tool", "name": "print_json"}

    response = await client.messages.create(**kwargs)

    return response


"""
This script provides functions for creating and validating dynamic models using Pydantic.
It supports advanced use cases, such as handling generic types, validating field choices,
and dynamically generating models based on function argument specifications.

Usage:

1. Creating a dynamic model from a function argument specification:
    >>> import inspect
    >>> def my_function(a: int, b: str, c: List[int]):
    ...     pass
    >>> argspec = inspect.getfullargspec(my_function)
    >>> model = create_model_from_argspec('MyModel', argspec)
    >>> print(model.model_dump())

2. Generating OpenAPI schemas:
    >>> from typing import List, Annotated
    >>> schema, metadata = get_schema_from_annotation(List[int])
    >>> print(schema)  # {'type': 'array', 'items': {'type': 'integer'}}
"""
from collections import defaultdict
from collections.abc import Callable
from typing import (
    Any,
    Literal,
    NamedTuple,
    TypeVar,
    get_args,
    get_origin,
)

from annotated_types import Ge, Le, MaxLen, MinLen
from pydantic import (
    BaseModel,
    create_model,
    model_validator,
)
from pydantic.fields import FieldInfo

from . import Field, File, Message


def dump(value: Any, context: Any | None = None) -> Any:
    """Dumps all models that live within a nested structure of arbitrary depth."""
    if isinstance(value, Message): # Message is no longer a BaseModel.
        return {
            "role": value.role,
            "content": [dump(c, context=context) for c in value.content],
        }
    elif isinstance(value, BaseModel):
        # Handle BaseModel instances as we handle dictionaries.
        return {k: dump(v, context=context) for k, v in value.__dict__.items()}
    elif isinstance(value, dict):
        return {k: dump(v, context=context) for k, v in value.items()}
    elif isinstance(value, list):
        return [dump(v, context=context) for v in value]
    elif isinstance(value, File):
        return File.serialize(value, {"context": context})
    elif isinstance(value, Exception):
        return {
            "error_type": type(value).__name__,
            "message": str(value),
            # "traceback": traceback.format_exc()
        }
    return value


def get_base_type(annotation: type[Any]) -> type[Any]:
    """Function that returns the base type of a type with an arbitrary number of layers
    (e.g. Annotation[List[List[T]]] -> T)."""
    origin = get_origin(annotation)
    if origin is None:
        return annotation
    args = get_args(annotation)
    if args:
        return get_base_type(args[0])
    return origin


def create_choices_validator(choices: list[str | int | float]) -> Callable:
    """Factory function to create a Pydantic validator for a list of choices."""
    def validate_choice(value: str | int | float, _info) -> str | int | float:
        if value not in choices:
            raise ValueError(f"Choice {value} is not one of: {choices}.")
        return value
    return validate_choice


def create_generics_validator(fields_names: list[str]) -> Callable:
    """Factory function to create a Pydantic model validator that ensures all 
    model generics (base types) are of the same type."""
    def validate_generics(fields_values: dict[str, Any], _info) -> dict[str, Any]:
        field_generic_type = None

        for field_name in fields_names:
            if field_name not in fields_values:
                continue

            field_value = fields_values[field_name]
            field_value_type = type(field_value)

            if isinstance(field_value, list):
                # Skip validation for empty lists
                if len(field_value) == 0:
                    continue
                field_value_type = type(field_value[0])
                if not all(isinstance(field_value_i, field_value_type) for field_value_i in field_value):
                    raise ValueError(f"All elements in the list '{field_name}' must be of the same base type!")

            if field_generic_type is None:
                field_generic_type = field_value_type
            elif field_generic_type != field_value_type:
                raise ValueError(f"Fields {fields_names} must have the same base types!")

        return fields_values

    return validate_generics


def create_model_from_argspec(name: str, argspec: NamedTuple) -> BaseModel:
    """Create a dynamic pydantic model from the argspec of a function."""
    fields = {}
    validators = {}
    generics = defaultdict(list)

    defaults = {}
    if argspec.defaults:
        defaults = dict(zip(argspec.args[-len(argspec.defaults):], argspec.defaults, strict=True))

    for field_name in argspec.args:
        field_default = defaults.get(field_name, ...) # Pydantic will use ... to mark this field as required.
        if not isinstance(field_default, FieldInfo):
            field_default = Field(default=field_default)

        field_type = argspec.annotations.get(field_name, Any)

        # If base type is a generic, we need to make sure that all values passed as this generic are of the same type.
        field_base_type = get_base_type(field_type)
        if isinstance(field_base_type, TypeVar):
            generics[field_base_type].append(field_name)

        json_schema_extra = getattr(field_default, "json_schema_extra", None)
        if isinstance(json_schema_extra, dict) and "choices" in json_schema_extra:
            choices = json_schema_extra.pop("choices")
            field_type = Literal.__getitem__(tuple(choices))
            # choices_validator = create_choices_validator(choices)
            # validators[f"{field_name}_validator"] = field_validator(field_name)(choices_validator)

        fields[field_name] = (field_type, field_default)

    for generic, generic_fields_names in generics.items():
        generic_validator = create_generics_validator(generic_fields_names)
        validators[f"{generic}_validator"] = model_validator(mode="before")(generic_validator)
    
    model = create_model(name, __validators__=validators, **fields)
    return model


def issubclass_safe(candidate: Any, base_class: type[Any]) -> bool:
    """Helper function to avoid checking on annotations types always before calling issubclass."""
    if isinstance(candidate, type):
        return issubclass(candidate, base_class)
    else:
        return False


def create_model_from_annotation(name: str, annotation: type[Any] | None) -> BaseModel:
    """Generate an openapi compliant json schema from a function return annotation."""
    # If the type is already a BaseModel, simply return the schema.
    if issubclass_safe(annotation, BaseModel):
        return annotation

    if annotation is None:
        model = create_model(name, **{"return": (Literal[None], ...)})
    elif annotation is Ellipsis:
        model = create_model(name, **{"return": (Any, ...)})
    else:
        model = create_model(name, **{"return": (annotation, ...)})

    # If the annotation is Annotated, pydantic will store the metadata in the FieldInfos.
    
    return model


def create_model_from_fields(name: str, model_fields: dict[str, FieldInfo]) -> BaseModel:
    """Create a pydantic model from a dictionary of fields."""
    fields = {}
    validators = {}
    generics = defaultdict(list)

    for field_name, field_info in model_fields.items():
        field_type = field_info.annotation
        fields[field_name] = (field_type, field_info)

        # If base type is a generic, we need to make sure that all values passed as this generic are of the same type.
        field_base_type = get_base_type(field_type)
        if isinstance(field_base_type, TypeVar):
            generics[field_base_type].append(field_name)

        # No need to check for json schema extra choices here. Since we'll be getting fields from steps models.
    
    for generic, generic_fields_names in generics.items():
        generic_validator = create_generics_validator(generic_fields_names)
        validators[f"{generic}_validator"] = model_validator(mode="before")(generic_validator)
    
    model = create_model(name, __validators__=validators, **fields)
    return model


def merge_model_fields(*args: FieldInfo) -> FieldInfo:
    """Merge an arbitrary number of pydantic fields into a single one if possible.
    It enforces the most restrictive validators."""
    field_annotation = None
    field_metadata = []
    field_descriptions = []
    field_kwargs = {"default": None}

    for field_info in args:
        if field_annotation is None:
            field_annotation = field_info.annotation
        else:
            if field_annotation != field_info.annotation:
                raise ValueError("Cannot merge fields with different annotations.")
        # If any of the merged fields is required, the merged field is required.
        if field_info.is_required():
            field_kwargs.pop("default", None)
        # Collect all descriptions for later formatting.
        field_descriptions.append(field_info.description)
        # The remaining validators live inside metadata as common annotated types.
        for metadata in field_info.metadata:
            if hasattr(metadata, "pattern"):
                raise NotImplementedError("Regex merging is not implemented.")
            elif isinstance(metadata, Ge):
                current_ge = field_kwargs.get("ge", float("-inf"))
                current_le = field_kwargs.get("le", float("inf"))
                new_ge = metadata.ge
                if new_ge > current_le:
                    raise ValueError(f"Cannot merge fields with incompatible constraints: ge {new_ge} > le {current_le}.")
                field_kwargs["ge"] = max(current_ge, new_ge)
            elif isinstance(metadata, Le):
                current_le = field_kwargs.get("le", float("inf"))
                current_ge = field_kwargs.get("ge", float("-inf"))
                new_le = metadata.le
                if new_le < current_ge:
                    raise ValueError(f"Cannot merge fields with incompatible constraints: le {new_le} < ge {current_ge}.")
                field_kwargs["le"] = min(current_le, new_le)
            elif isinstance(metadata, MaxLen):
                current_max_length = field_kwargs.get("max_length", float("inf"))
                current_min_length = field_kwargs.get("min_length", 0)
                new_max_length = metadata.max_length
                if new_max_length < current_min_length:
                    raise ValueError(f"Cannot merge fields with incompatible constraints: max_length {new_max_length} < min_length {current_min_length}.")
                field_kwargs["max_length"] = min(current_max_length, new_max_length)
            elif isinstance(metadata, MinLen):
                current_min_length = field_kwargs.get("min_length", 0)
                current_max_length = field_kwargs.get("max_length", float("inf"))
                new_min_length = metadata.min_length
                if new_min_length > current_max_length:
                    raise ValueError(f"Cannot merge fields with incompatible constraints: min_length {new_min_length} > max_length {current_max_length}.")
                field_kwargs["min_length"] = max(current_min_length, new_min_length)
            else:
                field_metadata.append(metadata)

        # Merge choices with the intersection of the choices of all fields.
        if isinstance(field_info.json_schema_extra, dict):
            choices = field_info.json_schema_extra.get("choices", None)
            if choices is not None:
                if "choices" not in field_kwargs:
                    field_kwargs["choices"] = choices
                else:
                    current_choices = field_kwargs["choices"]
                    field_kwargs["choices"] = list(set(current_choices) & set(choices))

    # Check if any of these choices are in conflict with the other validators.
    if "choices" in field_kwargs:
        valid_choices = []
        ge = field_kwargs.get("ge", float("-inf"))
        le = field_kwargs.get("le", float("inf"))
        min_length = field_kwargs.get("min_length", 0)
        max_length = field_kwargs.get("max_length", float("inf"))
        for choice in field_kwargs["choices"]:
            if isinstance(choice, int | float):
                if ge <= choice <= le:
                    valid_choices.append(choice)
            elif isinstance(choice, str):
                if min_length <= len(choice) <= max_length:
                    valid_choices.append(choice)
            else:
                valid_choices.append(choice)
        field_kwargs["choices"] = valid_choices
        # After validating the choices, we can remove the remaining validators since they are less restrictive in nature.
        field_kwargs.pop("ge", None)
        field_kwargs.pop("le", None)
        field_kwargs.pop("min_length", None)
        field_kwargs.pop("max_length", None)

    # Merge descriptions.
    if len(field_descriptions):
        field_kwargs["description"] = "Merged field descriptions:"
        for field_description in field_descriptions:
            field_kwargs["description"] += f"\n- {field_description}"

    # Init field info, and manually add annotation and extra metadata.
    field_info = Field(**field_kwargs)
    field_info.annotation = field_annotation
    field_info.metadata.extend(field_metadata)
    return field_info



"""
This module provides a File class that wraps various types of file sources,
including local files, URLs, enabling uniform interaction through
a file-like interface.

It is inspired by and based on the design found in the Replicate Cog project.
See original implementation at:
https://github.com/replicate/cog/blob/main/python/cog/types.py

Usage:

1. Validating and opening a local file:
   >>> file_instance = File.validate('path/to/local/file.txt')
   >>> print(file_instance.read())

2. Validating and opening a data url:
   >>> file_instance = File.validate('data:text/plain;base64,SGVsbG8sIFdvcmxkIQ%3D%3D')
   >>> print(file_instance.read()) # b'Hello, World!'

3. Opening a file from a URL:
   >>> from PIL import Image
   >>> file_instance = File.validate('https://example.com/image.png')
   >>> image = Image.open(file_instance)
   >>> image.show()

4. Opening a file-like object:
   >>> from io import BytesIO
   >>> file_instance = File.validate(BytesIO(b'Hello, World!'))
   >>> print(file_instance.read())
"""

import base64
import io
import mimetypes
import os
from collections.abc import Callable, Iterator
from pathlib import Path
from typing import Any
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
from pydantic import (
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    SerializationInfo,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
)
from pydantic_core import CoreSchema, core_schema


class File(io.IOBase):
    """A wrapper class that provides a uniform file-like interface for various file sources.

    This class can handle local files, URLs, S3 URIs, and data URLs, providing a consistent
    interface for file operations regardless of the source. It implements the io.IOBase interface
    and provides lazy loading of file contents.

    Attributes:
        __source__: The original source identifier (path, URL, etc.)
        __source_scheme__: The scheme of the source (local_path, url, s3, data_url, bytes)
        __source_extension__: The file extension if available
        __fileobj__: The actual file object once loaded
        __fetcher__: Function to fetch/load the file content
    """

    __slots__ = ("__source__", "__source_scheme__", "__source_extension__", "__fileobj__", "__fetcher__")

    __max_file_size__: int | None = (
        int(os.getenv("TIMBAL_MAX_FILE_SIZE")) if os.getenv("TIMBAL_MAX_FILE_SIZE") else None
    )
    """ Maximum allowed file size in bytes"""


    def __init__(
        self,
        source: Any,
        source_scheme: str,
        source_extension: str | None = None,
        fetcher: Callable[[], io.IOBase] | None = None,
    ) -> None:
        object.__setattr__(self, "__source__", source)
        object.__setattr__(self, "__source_scheme__", source_scheme)
        object.__setattr__(self, "__source_extension__", source_extension)
        object.__setattr__(self, "__fetcher__", fetcher)
        if isinstance(source, io.IOBase):
            object.__setattr__(self, "__fileobj__", source)
        else:
            object.__setattr__(self, "__fileobj__", None)


    def __str__(self) -> str:
        return self.__source__


    def __repr__(self) -> str:
        return f"File(source={str(self)})"


    def __getattr__(self, name: str) -> Any:
        """Proxy attribute access through to the wrapped file object."""
        if name in ("__source__", "__source_scheme__", "__source_extension__", "__fileobj__", "__fetcher__"):
            raise AttributeError(name)
        else:
            return getattr(self.__wrapped__, name)


    def __setattr__(self, name: str, value: Any) -> None:
        """Proxy attribute assignment through to the wrapped file object."""
        if hasattr(type(self), name):
            object.__setattr__(self, name, value)
        else:
            setattr(self.__wrapped__, name, value)


    def __delattr__(self, name: str) -> None:
        """Proxy attribute deletion through to the wrapped file object."""
        if hasattr(type(self), name):
            object.__delattr__(self, name)
        else:
            delattr(self.__wrapped__, name)


    def __iter__(self) -> Iterator[bytes]:
        """Iterate over the wrapped file object."""
        return iter(self.__wrapped__)
    

    @property
    def __wrapped__(self) -> Any:
        """Get the underlying file object, fetching it if necessary."""
        fileobj = object.__getattribute__(self, "__fileobj__")
        if fileobj is None:
            fetcher = object.__getattribute__(self, "__fetcher__")
            if fetcher is None:
                raise ValueError("File object is not properly initialized.")
            fileobj = fetcher()
            object.__setattr__(self, "__fileobj__", fileobj)
        return fileobj


    # Overwrite some io.IOBase methods to proxy to the wrapped file object (explicitly).
    def readable(self) -> bool:
        return self.__getattr__("readable")()


    def writable(self) -> bool:
        return self.__getattr__("writable")()


    def seekable(self) -> bool:
        return self.__getattr__("seekable")()


    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.__getattr__("seek")(offset, whence)


    @classmethod
    def validate(
        cls, 
        value: ValidatorFunctionWrapHandler, 
        info: dict | ValidationInfo | None = None, # noqa: ARG003
    ) -> "File":
        """Create a new Field instance validating a local path, an url, an s3 uri, a data url or a file-like object.
        Validation info can be used to pass context information to the file fetcher:
            >>> model_instance = model.model_validate(
            ...     {**model_params_dict},
            ...     context={...},
            ... )
        """
        if isinstance(value, cls):
            return value

        if isinstance(value, bytes | bytearray):
            return File(io.BytesIO(value), source_scheme="bytes")

        if isinstance(value, io.IOBase):
            return File(value, source_scheme="bytes")

        if isinstance(value, Path):
            value = value.expanduser().resolve().as_posix()

        if not isinstance(value, str):
            raise ValueError("File must be a string, path, or file-like object.")

        parsed_url = urlparse(value)

        # Check for valid local path.
        if parsed_url.scheme == "":
            path = Path(value).expanduser().resolve()
            if not path.exists() or not path.is_file():
                raise ValueError("Invalid file path.")
            if cls.__max_file_size__ is not None:
                file_size = path.stat().st_size  # bytes
                if file_size > cls.__max_file_size__:
                    raise ValueError(f"File size exceeds the maximum allowed size of {cls.__max_file_size__} bytes.")
            parsed_url_extension = path.suffix
            return File(
                value,
                source_scheme="local_path",
                source_extension=parsed_url_extension,
                fetcher=lambda: cls._fetch_local_file(path=path),
            )

        # Data urls: data:[<mediatype>][;base64],<data>.
        if parsed_url.scheme == "data":
            if cls.__max_file_size__ is not None:
                comma_idx = value.find(",")
                # Redundant check.
                if comma_idx == -1:
                    raise ValueError("Invalid data URL.")
                encoded = value[comma_idx + 1 :]
                # ? We could count the padding if we want to be more precise.
                estimated_size = len(encoded) * 3 // 4
                if estimated_size > cls.__max_file_size__:
                    raise ValueError(f"File size exceeds the maximum allowed size of {cls.__max_file_size__} bytes.")
            parsed_url_mimetype = parsed_url.path.split(";", 1)
            if len(parsed_url_mimetype) != 2:
                raise ValueError("Data url must specify a mimetype.")
            parsed_url_mimetype = parsed_url_mimetype[0]
            parsed_url_extension = mimetypes.guess_extension(parsed_url_mimetype)
            if parsed_url_extension is None:
                # Custom mimetypes must be specified as ^timbal\/\w*$
                if not parsed_url_mimetype.startswith("timbal/"):
                    raise ValueError("Custom mimetypes must be specified as 'timbal/<extension>'.")
                parsed_url_extension = "." + parsed_url_mimetype[7:]
            return File(
                value,
                source_scheme="data_url",
                source_extension=parsed_url_extension,
                fetcher=lambda: cls._fetch_data_url_file(data_url=value),
            )

        if parsed_url.scheme == "http" or parsed_url.scheme == "https":
            response = requests.head(value)
            if cls.__max_file_size__ is not None:
                content_length = int(response.headers.get("Content-Length", 0))
                if content_length > cls.__max_file_size__:
                    raise ValueError(f"File size exceeds the maximum allowed size of {cls.__max_file_size__} bytes.")
            parsed_url_name = parsed_url.path.split("/")[-1]
            parsed_url_extension = f".{parsed_url_name.split('.')[-1]}" if "." in parsed_url_name else ""
            return File(
                value,
                source_scheme="url",
                source_extension=parsed_url_extension,
                fetcher=lambda: cls._fetch_http_file(url=value),
            )

        raise ValueError("Invalid file source. Must be a local file path, data URL, or HTTP/HTTPS url.")


    @classmethod
    def serialize(
        cls, 
        value: Any, 
        info: dict | SerializationInfo | None = None, # noqa: ARG003
    ) -> str:
        """Serialize the file to a data url string. Bytes-like files are not supported will have octet-stream as content type."""
        # When creating a model with fields with File type that are nullable,
        # pydantic will pass None as the value to File.serialize.
        if value is None:
            return None

        if not isinstance(value, cls):
            raise ValueError("Cannot serialize a non-file object.")

        if value.__source_scheme__ == "data_url":
            return str(value)

        # We cannot safely serialize bytes-like files.
        if value.__source_scheme__ == "bytes":
            content_type = "application/octet-stream"
        else:
            content_type, _ = mimetypes.guess_type(str(value))
            if content_type is None:
                content_type = f"timbal/{value.__source_extension__}"

        # Ensure the file obj has the pointer at the start of the file.
        current_position = value.tell()
        if current_position != 0:
            value.seek(0)

        content = value.read()
        bs64_content = base64.b64encode(content).decode("utf-8")
        return f"data:{content_type};base64,{bs64_content}"


    @classmethod
    def _fetch_local_file(cls, path: Path) -> io.IOBase:
        """Fetch a local file from a valid path in the system."""
        return path.open("rb")


    @classmethod
    def _fetch_data_url_file(cls, data_url: str) -> io.IOBase:
        """Parse a data URL and return a file-like object."""
        res = urlopen(data_url)  # noqa: S310
        return io.BytesIO(res.read())


    @classmethod
    def _fetch_http_file(cls, url: str) -> io.IOBase:
        """Fetch a file from a HTTP/HTTPS URL."""
        res = requests.get(url, stream=True)
        res.raise_for_status()
        return io.BytesIO(res.content)


    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema: CoreSchema, _handler: GetJsonSchemaHandler) -> dict[str, Any]:
        """Defines what this type should be in openapi.json."""
        # https://docs.pydantic.dev/2.8/errors/usage_errors/#custom-json-schema
        json_schema = {
            "type": "string",
            "format": "uri",
            "description": "A file reference which can be a local path, a URL, or a data URL.",
        }
        return json_schema


    @classmethod
    def __get_pydantic_core_schema__(cls, _source: type[Any], _handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Defines how to serialize this type in the core schema."""
        return core_schema.with_info_plain_validator_function(
            cls.validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.serialize,
                info_arg=True,
                when_used="always",
            ),
        )



# ruff: noqa: F401
from .chat.content import Content, FileContent, TextContent, ToolResultContent, ToolUseContent
from .chat.tool import Tool
from .field import Field
from .file import File
from .message import Message



from timbal import Flow


def identity_handler(x):
    return x


def test_compile():
    flow = (
        Flow(id="test_compile")
        .add_step("1", identity_handler)
        .add_step("2", identity_handler)
        .add_link("1", "2")
    )

    flow.compile()

    assert hasattr(flow, '_dag')
    assert hasattr(flow, '_rev_dag')

    assert hasattr(flow, '_params_model')
    assert hasattr(flow, '_return_model')
    assert hasattr(flow, '_params_model_schema')
    assert hasattr(flow, '_return_model_schema')
    


import os
from typing import Any, Literal

import structlog
from dotenv import load_dotenv
from openai import AsyncOpenAI

from ...types import Field, Message, Tool

logger = structlog.getLogger("timbal.steps.llms.togetherai_llm")

load_dotenv()

async def handler(
    # This is not actually used inside the function. It's a hack to have it as an argument
    # so we can use it when mapping data in the flow.
    prompt: Message = Field(default=None, description="Message to send to the LLM."), # noqa: ARG001
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    memory: list[Message] = Field(description="A list containing the conversation history between the user and the LLM."),
    model: Literal[
        "deepseek-ai/DeepSeek-R1",
        "deepseek-ai/DeepSeek-V3",
        "meta-llama/Llama-3.3-70B-Instruct-Turbo",
        "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
        "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo",
        "meta-llama/Llama-3.2-3B-Instruct-Turbo",
        "Qwen/Qwen2.5-Coder-32B-Instruct",
        "Qwen/Qwen2-VL-72B-Instruct",
        "mistralai/Mistral-Small-24B-Instruct-2501",
        "mistralai/Mistral-7B-Instruct-v0.3",
        "mistralai/Mixtral-8x22B-Instruct-v0.1",
        "meta-llama/Llama-3.2-11B-Vision-Instruct-Turbo",
    ] | str = Field(default="meta-llama/Llama-3.2-3B-Instruct-Turbo", description="Name of the LLM model to use"),
    tools: list[Tool | dict] = Field(default=None, description="List of tools/functions that the LLM can call"),
    tool_choice: dict[str, Any] | str = Field(default = {"type": "auto"}, description="How the model should use the provided tools"),
    max_tokens: int = Field(default=None, description="Maximum length of response in tokens"),
    frequency_penalty: float = Field(default=0, description="Penalizes token repetition on a scale of -2.0 to 2.0. Higher values reduce repetition"),
    top_logprobs: int = Field(default=None, description="Number of most likely tokens (0-20) to return with probabilities"),
    presence_penalty: float = Field(default=0, description="Penalizes token presence on a scale of -2.0 to 2.0. Higher values encourage new topics"),
    seed: int = Field(default=None, description="Seed for deterministic sampling (Beta feature)"),
    stop: str | list[str] = Field(default=None, description="Up to 4 sequences where the model will stop generating further tokens."),
    temperature: float = Field(default=1, description="Amount of randomness injected into the response. Ranges from 0.0 to 2.0."),
    top_p: float = Field(default=1, description="Nucleus sampling parameter. Ranges from 0.0 to 1.0."),
    parallel_tool_calls: bool = Field(default=True, description="Whether to enable parallel function calling during tool use"),
    json_schema: dict = Field(default=None, description="The JSON input schema that the model must output."),
) -> Any:
    """This handler manages interactions with TogetherAI's API, processing requests with
    appropriate parameters and returning streaming responses.

    Args:
        prompt: Hack to pass an input to the LLM.
        memory: Chat history containing user and LLM messages.
        system_prompt: Instructions for the LLM to follow.
        model: Name of the TogetherAI model to use.
        tools: List of tools/functions the LLM can call.
        tool_choice: How the model should use the provided tools.
        max_tokens: Maximum number of tokens in the response.
        frequency_penalty: Value between -2.0 and 2.0 to penalize token frequency.
        top_logprobs: Number of most likely tokens to return (0-20). Requires logprobs=True.
        presence_penalty: Value between -2.0 and 2.0 to penalize token presence.
        seed: Seed for deterministic sampling (Beta feature).
        stop: Sequences where token generation should stop.
        temperature: Sampling temperature between 0 and 2.
        top_p: Nucleus sampling probability threshold.
        parallel_tool_calls: Whether to call multiple tools in parallel or not.
        json_schema: JSON schema for structured output.

    Returns:
        Any: The streaming response from the TogetherAI API.
    """
    client = AsyncOpenAI(
            api_key=os.environ.get("TOGETHER_API_KEY"),
            base_url="https://api.together.xyz/v1",
        )
    
    # Get default values for parameters
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    model = model.default if hasattr(model, 'default') else model
    tools = tools.default if hasattr(tools, 'default') else tools
    tool_choice = tool_choice.default if hasattr(tool_choice, 'default') else tool_choice
    max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    frequency_penalty = frequency_penalty.default if hasattr(frequency_penalty, 'default') else frequency_penalty
    top_logprobs = top_logprobs.default if hasattr(top_logprobs, 'default') else top_logprobs
    presence_penalty = presence_penalty.default if hasattr(presence_penalty, 'default') else presence_penalty
    seed = seed.default if hasattr(seed, 'default') else seed
    stop = stop.default if hasattr(stop, 'default') else stop
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    parallel_tool_calls = parallel_tool_calls.default if hasattr(parallel_tool_calls, 'default') else parallel_tool_calls
    json_schema = json_schema.default if hasattr(json_schema, 'default') else json_schema

    if model in ["deepseek-ai/DeepSeek-R1", "deepseek-ai/DeepSeek-V3", "meta-llama/Llama-3.3-70B-Instruct-Turbo", "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo", 
                 "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo", "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo", "meta-llama/Llama-3.2-3B-Instruct-Turbo", 
                 "Qwen/Qwen2.5-Coder-32B-Instruct", "mistralai/Mistral-Small-24B-Instruct-2501", "mistralai/Mistral-7B-Instruct-v0.3",
                  "mistralai/Mixtral-8x22B-Instruct-v0.1"]:
        logger.warning("Not multimodal LLM. This model does not support image type inputs.", model=model)
    
    messages = []
    if system_prompt:
        messages.append({"role": "system", "content": system_prompt})
    for msg in memory:
        openai_msg = msg.to_openai_input()
        # Extract content from the message
        if isinstance(openai_msg["content"], list):
            is_text_only = all(
                (isinstance(item, dict) and "text" in item) or isinstance(item, str)
                for item in openai_msg["content"]
            )
            if is_text_only:
                content_text = []
                for item in openai_msg["content"]:
                    if isinstance(item, dict) and "text" in item:
                        content_text.append(item["text"])
                    elif isinstance(item, str):
                        content_text.append(item)
                openai_msg["content"] = " ".join(content_text)
            messages.append(openai_msg)

    kwargs = {
        "model": model,
        "messages": messages,
        "stream": True,
        "logprobs": top_logprobs,
        "presence_penalty": presence_penalty,
        "temperature": temperature,
        "top_p": top_p,
    }

    # Optional parameters
    if frequency_penalty:
        kwargs["frequency_penalty"] = frequency_penalty
    if stop:
        kwargs["stop"] = stop
    if seed:
        kwargs["seed"] = seed
    if max_tokens:
        kwargs["max_tokens"] = max_tokens

    # Change parameters if provided to TogetherAI format
    if tools:
        tools_openai = []
        for tool in tools:
            if isinstance(tool, Tool):
                tools_openai.append(tool.to_openai())
            else:
                tools_openai.append(tool)
        kwargs["tools"] = tools_openai
        if isinstance(tool_choice, dict):
            if tool_choice["type"] != "function":
                tool_choice = tool_choice["type"]
        kwargs["tool_choice"] = tool_choice
        kwargs["parallel_tool_calls"] = parallel_tool_calls

    if json_schema:
        kwargs["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "name": "print_json",
                "schema": {
                    "type": "object",
                    "properties": json_schema
                }
            }
        }

    response = await client.chat.completions.create(**kwargs)

    return response


from enum import Enum

from pydantic import BaseModel, ConfigDict


class TimbalPlatformAuthType(str, Enum):
    BEARER = "bearer"
    CUSTOM = "custom"


class TimbalPlatformAuthConfig(BaseModel):
    """Configuration for platform authentication.
    At the moment, supports bearer tokens and custom headers.
    """

    type: TimbalPlatformAuthType
    """Type of authentication to use."""
    token: str
    """Token included in the authentication header."""
    header_name: str | None = None
    """If type is CUSTOM, this will be the name of the header to use."""

    @property
    def header_key(self) -> str:
        """Format header key for requests authenticating with the platform."""
        if self.type == TimbalPlatformAuthType.BEARER:
            return "Authorization"
        elif self.type == TimbalPlatformAuthType.CUSTOM:
            return self.header_name
        else:
            raise NotImplementedError(f"Unknown auth type: {self.type}")

    @property 
    def header_value(self) -> str:
        """Format header value for requests authenticating with the platform."""
        if self.type == TimbalPlatformAuthType.BEARER:
            return f"Bearer {self.token}"
        elif self.type == TimbalPlatformAuthType.CUSTOM:
            return self.token
        else:
            raise NotImplementedError(f"Unknown auth type: {self.type}")


class TimbalPlatformAppConfig(BaseModel):
    """Application-specific platform configuration.
    Contains identifiers to the platform resource the run context applies to.
    """

    org_id: str 
    """Organization identifier."""
    app_id: str
    """Application identifier."""
    version_id: str
    """Application version identifier."""


class TimbalPlatformConfig(BaseModel):
    """Complete platform configuration.
    Contains all the information needed to authenticate and identify the platform resource the run context applies to.
    """

    host: str
    """Platform host."""
    auth_config: TimbalPlatformAuthConfig
    """Platform authentication configuration."""
    app_config: TimbalPlatformAppConfig
    """Platform application configuration."""


class RunContext(BaseModel):
    """Context for a run.
    This is shared between all steps in a flow (including nested subflows).
    """
    # Allow for extra fields.
    model_config = ConfigDict(extra="allow")

    id: str | None = None
    """Unique identifier for the run.
    We allow for the id to be None so the dev has more flexibility when trying things out.
    The existance of this field should be enforced when creating a Snapshot and using the state saver.
    """
    parent_id: str | None = None
    """Whether this run is a direct child of another run.
    Can be used to recursively retrieve all runs into a single list -> chat history.
    Can also be used to create a new branch from a specific run -> rewind.
    """
    timbal_platform_config: TimbalPlatformConfig | None = None
    """Platform configuration for the run."""



import importlib.util
import socket
from pathlib import Path
from typing import Any

from pydantic import BaseModel


class ModuleSpec(BaseModel):
    path: Path
    object_name: str | None = None


def load_module(module_spec: ModuleSpec) -> Any:
    """Load an object from a module."""
    path = module_spec.path 
    object_name = module_spec.object_name

    spec = importlib.util.spec_from_file_location(path.stem, path.as_posix())
    if spec and spec.loader:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        if object_name:
            if hasattr(module, object_name):
                obj = getattr(module, object_name)
                return obj
            else:
                raise ValueError(f"Module {path} has no object {object_name}")
        else:
            raise NotImplementedError("? support loading entire module")
    else:
        raise ValueError(f"Failed to load module {path}")


def is_port_in_use(port: int) -> bool:
    """Check if a TCP port is currently in use on localhost."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex(("localhost", port)) == 0



from typing import Any

import requests
import structlog
from pydantic import TypeAdapter

from ...types.models import dump
from ..context import RunContext
from ..data import Data
from ..snapshot import Snapshot
from .base import BaseSaver

logger = structlog.get_logger("timbal.state.savers.timbal_platform")


class TimbalPlatformSaver(BaseSaver):
    """A state saver that stores snapshots in the Timbal platform.

    This state saver is used to store snapshots in the Timbal platform.
    You can see the logs and snapshots history from the platform UI.

    Note:
        This state saver requires a `TimbalPlatformConfig` to be passed within the `RunContext`.
    """
    _get_warning_shown = False
    _put_warning_shown = False


    @staticmethod
    def _load_snapshot_from_res_body(res_body: dict[str, Any]) -> Snapshot:
        res_body["data"] = {
            k: TypeAdapter(Data).validate_python(v)
            for k, v in res_body["data"].items()
        }
        return Snapshot(**res_body)


    def get_last(
        self,
        path: str,
        context: RunContext,
    ) -> Snapshot | None:
        """See base class."""
        if not context.timbal_platform_config:
            if not self._get_warning_shown:
                logger.warning(
                    "TimbalPlatformSaver: Missing config for GET operation. " \
                    "Pass config to the RunContext to enable fetching snapshots from the platform. " \
                    "You can safely ignore this warning if you intend to push this app to the platform later."
                )
                self._get_warning_shown = True
            return None

        if context.parent_id is None:
            return None

        # No need to check for anything else, the timbal platform config will already be validated.

        host = context.timbal_platform_config.host

        auth_config = context.timbal_platform_config.auth_config
        headers = {auth_config.header_key: auth_config.header_value}

        app_config = context.timbal_platform_config.app_config
        org_id = app_config.org_id
        app_id = app_config.app_id
        resource_path = f"orgs/{org_id}/apps/{app_id}/runs/{context.parent_id}"

        res = requests.get(
            f"https://{host}/{resource_path}/snapshots", 
            headers=headers,
            params={"path": path},
        )
        res.raise_for_status()

        res_body = res.json()
        return self._load_snapshot_from_res_body(res_body)


    def put(
        self, 
        snapshot: Snapshot,
        context: RunContext,
    ) -> None:
        """See base class."""
        if not context.timbal_platform_config:
            if not self._put_warning_shown:
                logger.warning(
                    "TimbalPlatformSaver: Missing config for PUT operation. " \
                    "Pass config to the RunContext to enable storing snapshots on the platform. " \
                    "You can safely ignore this warning if you intend to push this app to the platform later."
                )
                self._put_warning_shown = True
            return None

        # No need to check for anything else, the timbal platform config will already be validated.

        host = context.timbal_platform_config.host

        auth_config = context.timbal_platform_config.auth_config
        headers = {auth_config.header_key: auth_config.header_value}

        app_config = context.timbal_platform_config.app_config
        org_id = app_config.org_id
        app_id = app_config.app_id
        resource_path = f"orgs/{org_id}/apps/{app_id}/runs/{context.id}"

        body = dump(snapshot)

        res = requests.post(
            f"https://{host}/{resource_path}/snapshots", 
            headers=headers,
            json=body,
        )
        res.raise_for_status()
    



from typing import Any

from pydantic import BaseModel


class Tool(BaseModel):
    """
    A class representing a tool or function that can be used by AI models.
    """
    name: str 
    """The name of the tool."""
    description: str 
    """A description of the tool."""
    input_schema: dict[str, Any]
    """The input schema for the tool."""

    def to_openai(self) -> dict[str, Any]:
        """Convert the tool to OpenAI's expected format."""
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.input_schema,
            }
        }

    def to_anthropic(self) -> dict[str, Any]:
        """Convert the tool to Anthropic's expected format."""
        return {
            "name": self.name,
            "description": self.description,
            "input_schema": self.input_schema,
        }



import pytest
from pydantic import TypeAdapter
from timbal.errors import DataKeyError
from timbal.state.data import Data, DataError, DataMap, DataValue, get_data_key


def test_data_value():
    data = TypeAdapter(Data).validate_python({
        "type": "value", 
        "value": "test",
    })

    assert isinstance(data, DataValue)
    assert data.resolve() == "test"


def test_data_error():
    data = TypeAdapter(Data).validate_python({
        "type": "error", 
        "error": "test",
    })

    assert isinstance(data, DataError)
    assert data.resolve() == data


def test_data_map():
    data = TypeAdapter(Data).validate_python({
        "type": "map", 
        "key": "1.return"
    })

    assert isinstance(data, DataMap)

    context_data = {"1.return": DataValue(value="test")}
    assert data.resolve(context_data=context_data) == "test"

    with pytest.raises(DataKeyError):
        data = DataMap(key="abc.return")
        print(data.resolve(context_data=context_data)) # noqa: T201


def test_get_data_key():
    data = {"1.return": DataValue(value=[{"test_key": ["test_value"]}])}

    assert get_data_key(data, "1.return.0.test_key.0") == "test_value"

    # Accessing a list index that doesn't exist
    with pytest.raises(KeyError):
        get_data_key(data, "1.return.0.test_key.1")

    # Accessing a list index from a dict
    with pytest.raises(KeyError):
        get_data_key(data, "1.return.0.0")

    # Treating a list as an object with properties
    with pytest.raises(KeyError):
        get_data_key(data, "1.return.0.test_key.something")

    # Searching for a key that doesn't exist
    with pytest.raises(DataKeyError):
        get_data_key(data, "1.metadata")

    # Regex validation
    with pytest.raises(ValueError):
        get_data_key(data, "invalid data key")



from ..context import RunContext
from ..snapshot import Snapshot
from .base import BaseSaver


class InMemorySaver(BaseSaver):
    """An in-memory state saver.

    This state saver stores snapshots in memory using a python in-memory list.

    Note:
        Only use `InMemorySaver` for debugging or testing purposes.
        For production use cases, use a persistent state saver like `PostgresSaver`.
    """

    def __init__(
        self, 
        snapshots: list[Snapshot] | None = None,
    ) -> None:
        """Initialize an InMemorySaver instance.

        Args:
            snapshots: Optional list of Snapshot objects to initialize the saver with.
                       Defaults to an empty list.
        """
        self.snapshots = snapshots if snapshots is not None else []
    

    def get_last(
        self, 
        path: str,
        context: RunContext,
    ) -> Snapshot | None:
        """See base class."""
        if context.parent_id is None:
            return None

        for snapshot in self.snapshots[::-1]:
            if snapshot.path == path and snapshot.id == context.parent_id:
                return snapshot

        return None
    

    def put(
        self, 
        snapshot: Snapshot,
        context: RunContext, # noqa: ARG002
    ) -> None:
        """See base class."""
        # Since we're appending snapshots to a python list and there's no intrinsic way of ensuring
        # unicity of ids, we need to check if the snapshot already exists.
        for snapshot_i in reversed(self.snapshots):
            if snapshot_i.id == snapshot.id and snapshot_i.path == snapshot.path:
                raise ValueError(f"Snapshot with id {snapshot.id} and path {snapshot.path} already exists.")

        self.snapshots.append(snapshot)



import argparse
import asyncio
import os
import signal
import sys
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from pathlib import Path

import structlog
import uvicorn
from dotenv import load_dotenv
from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse

from .. import __version__
from ..logs import setup_logging
from ..state import RunContext
from ..types.models import dump
from .utils import ModuleSpec, is_port_in_use, load_module

logger = structlog.get_logger("timbal.server.http")


@asynccontextmanager
async def lifespan(
    app: FastAPI,
    module_spec: ModuleSpec,
) -> AsyncGenerator[None, None]:
    """Manages the lifecycle of the FastAPI application.

    This context manager handles the setup and teardown of the application,
    including loading the specified Python module and its object.

    Args:
        app: The FastAPI application instance.
        module_spec: A tuple containing (Path to Python module, Optional object name).
                     If object name is provided, it will be loaded from the module
                     and set as app.state.flow.

    Raises:
        ValueError: If the module or specified object cannot be loaded.
        NotImplementedError: If no object name is specified (currently unsupported).
    """
    # ? Any additional setup

    logger.info("loading_module", module_spec=module_spec)
    
    app.state.flow = load_module(module_spec)
    
    yield
    
    # ? Any additional cleanup


def create_app(
    module_spec: ModuleSpec, 
    shutdown_event: asyncio.Event,
) -> FastAPI:
    """HTTP server implementation for Timbal.

    This module provides the HTTP server implementation for running Timbal flows over a REST API.
    It handles module loading, parameter validation, and flow execution.

    Args:
        module_spec: Path to the flow module and object name.
        shutdown_event: Event to signal shutdown.
    """
    app = FastAPI(lifespan=lambda app: lifespan(app, module_spec))

    @app.get("/healthcheck")
    async def healthcheck() -> Response:
        return Response(status_code=204)

    @app.post("/shutdown")
    async def shutdown() -> Response:
        shutdown_event.set()
        return Response(status_code=204)

    @app.get("/params_model_schema")
    async def params_model_schema() -> Response:
        params_model_schema = app.state.flow.params_model_schema()
        return JSONResponse(
            status_code=200,
            content=params_model_schema,
        )

    @app.get("/return_model_schema")
    async def return_model_schema() -> Response:
        return_model_schema = app.state.flow.return_model_schema()
        return JSONResponse(
            status_code=200,
            content=return_model_schema,
        )

    @app.post("/run")
    async def run(req: Request) -> Response:
        req_data = await req.json()
        req_context = req_data.pop("context", None)
        req_context = RunContext.model_validate(req_context)
        res_content = await app.state.flow.complete(context=req_context, **req_data)
        res_content = dump(res_content)
        return JSONResponse(
            status_code=200,
            content=res_content,
        )

    return app


async def main(
    host: str,
    port: int,
    workers: int,
    module_spec: ModuleSpec,
) -> None:
    """Runs the HTTP server with the specified configuration.

    Sets up a FastAPI application with healthcheck, shutdown, and run endpoints.
    Handles graceful shutdown on SIGTERM and SIGINT signals.

    Args:
        host: The hostname to bind the server to.
        port: The port number to listen on.
        workers: Number of worker processes to spawn.
        module_spec: A tuple containing (Path to Python module, Optional object name).
                     If object name is provided, it will be loaded from the module
                     and set as app.state.flow.
    """
    shutdown_event = asyncio.Event()

    app = create_app(module_spec, shutdown_event)

    config = uvicorn.Config(
        app, 
        host=host, 
        port=port,
        workers=workers,
        log_level=None,
    )
    server = uvicorn.Server(config)
    
    def signal_handler() -> None:
        logger.info("shutdown_signal_received")
        shutdown_event.set()

    for sig in (signal.SIGTERM, signal.SIGINT):
        signal.signal(sig, lambda _signum, _frame: signal_handler())

    serve_task = asyncio.create_task(server.serve())
    await shutdown_event.wait()
    server.should_exit = True
    await serve_task


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Timbal HTTP server.")
    parser.add_argument(
        "-v", 
        "--version", 
        action="store_true", 
        help="Show version and exit."
    )
    parser.add_argument(
        "--module_spec",
        dest="module_spec",
        type=str,
        help="Path to a python module and optional object (format: path/to/file.py::object_name)",
    )
    parser.add_argument(
        "--host",
        dest="host",
        type=str,
        default="0.0.0.0",
        help="Host to bind to.",
    )
    parser.add_argument(
        "--port",
        dest="port",
        type=int,
        default=4444,
        help="Port to bind to.",
    )
    parser.add_argument(
        "--workers",
        dest="workers",
        type=int,
        default=1,
        help="Number of worker processes. Defaults to number of CPUs, or 1 if using a GPU.",
    )
    args = parser.parse_args()

    if args.version:
        print(f"timbal.servers.http {__version__}") # noqa: T201
        sys.exit(0)

    # We can overwrite the env TIMBAL_FLOW variable with the --module_spec flag.
    module_spec = args.module_spec
    if not module_spec:
        module_spec = os.getenv("TIMBAL_FLOW")

    if not module_spec:
        print("No module spec provided. Set TIMBAL_FLOW env variable or use --module_spec to specify a module to load.") # noqa: T201
        sys.exit(1)

    module_parts = module_spec.split(":")
    if len(module_parts) > 2:
        print("Invalid module spec format. Use 'path/to/file.py:object_name' or 'path/to/file.py'") # noqa: T201
        sys.exit(1)
    elif len(module_parts) == 2:
        module_path, module_name = module_parts
        module_spec = ModuleSpec(
            path=Path(module_path).expanduser().resolve(), 
            object_name=module_name,
        )
    else:
        module_spec = ModuleSpec(
            path=Path(module_parts[0]).expanduser().resolve(), 
            object_name=None,
        )

    if is_port_in_use(args.port):
        print(f"Port {args.port} is already in use. Please use a different port.") # noqa: T201
        sys.exit(1)

    load_dotenv()
    setup_logging()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        loop.run_until_complete(main(
            host=args.host, 
            port=args.port,
            workers=args.workers,
            module_spec=module_spec,
        ))
    except Exception as e:
        logger.error("server_stopped", error=str(e))
    finally:
        pending = asyncio.all_tasks(loop)
        logger.info("loop_pending_tasks", count=len(pending))
        if pending:
            loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
        loop.close()
        logger.info("loop_closed")



from typing import Any


class TimbalError(Exception):
    """Base class for all Timbal errors."""


class DataKeyError(TimbalError):
    """Error raised when a data key is not found."""


class StepKeyError(TimbalError):
    """Error raised when trying to set a step param that does not exist."""


class InvalidLinkError(TimbalError):
    """Error raised when trying to add a link between two steps that is invalid 
    (for some reason specified in the message)."""


class StepExecutionError(TimbalError):
    """Error raised when an error occurs during the execution of a step."""
    def __init__(self, input: Any, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.input = input


class FlowExecutionError(TimbalError):
    """Error raised when a step of the flow fails for some reason."""



from typing import Any

import pytest
from timbal import Flow
from timbal.errors import StepKeyError
from timbal.graph import Step
from timbal.state.data import DataMap
from timbal.types import Field


def identity_handler(x: Any = Field(default=None)) -> Any:
    return x


def identity_handler_2(x: Any = Field(default=None)) -> Any:
    return x


@pytest.mark.asyncio
async def test_add_step_with_handler():
    flow = Flow(id="test_add_step_with_handler")

    flow.add_step("step_1", identity_handler)
    assert "step_1.x" not in flow.data

    flow.add_step("step_2", identity_handler, x=1)
    assert flow.data["step_2.x"].resolve() == 1

    with pytest.raises(StepKeyError):
        flow.add_step("step_3", identity_handler, y=1)


@pytest.mark.asyncio
async def test_add_step_with_step():
    flow = Flow(id="test_add_step_with_step")

    step = Step(id="this_will_be_overriden", handler_fn=identity_handler)
    flow.add_step("step_1", step)

    assert "step_1" in flow.steps
    assert "this_will_be_overriden" not in flow.steps


@pytest.mark.asyncio
async def test_add_step_override():
    flow = Flow(id="test_add_step_override")

    flow.add_step("step_1", identity_handler)

    with pytest.raises(ValueError):
        flow.add_step("step_1", identity_handler)


@pytest.mark.asyncio
async def test_add_step_without_id():
    flow = Flow(id="test_add_step_without_id")

    flow.add_step(identity_handler, x=1)

    assert "identity_handler" in flow.steps
    assert flow.data["identity_handler.x"].resolve() == 1


@pytest.mark.asyncio
async def test_automatic_link_creation():
    flow = (
        Flow(id="test_automatic_link_creation")
        .add_step("step_1", identity_handler, x=1)
        .add_step("step_2", identity_handler_2)
        .set_data_map("step_2.x", "step_1.return")
        .set_output("step_2.return", "result")
    )

    result = await flow.complete()

    assert len(flow.links) == 1 and flow.links["step_1-step_2"]
    assert "step_2" in flow.steps
    assert result.output["result"] == 1


@pytest.mark.asyncio
async def test_add_step_data_map():
    flow = (
        Flow(id="test_add_step_data_map")
        .add_step("step_1", identity_handler, x=1)
        .add_step("step_2", identity_handler_2, x=DataMap(key="step_1.return"))
        .set_output("step_2.return", "result")
    )

    result = await flow.complete()

    assert len(flow.links) == 1 and flow.links["step_1-step_2"]
    assert "step_2" in flow.steps
    assert result.output["result"] == 1



import pytest
from timbal import Flow
from timbal.types import Message


@pytest.mark.asyncio
async def test_complete():
    flow = (
        Flow(id="flow")
        .add_llm(id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "response")
    )

    flow_output_event = await flow.complete(prompt="What is the capital of France?")
    flow_output = flow_output_event.output

    assert "response" in flow_output and isinstance(flow_output["response"], Message)



"""Field wrapper around Pydantic's Field, adapted for Replicate's Cog field specification.
Enables model parameter definitions with validation rules including required/optional fields,
numeric bounds, string constraints, regex patterns, and enumerated choices.

See Also:
    - Pydantic Field: https://docs.pydantic.dev/latest/api/fields/
    - Cog Specification: https://github.com/replicate/cog/blob/main/python/cog/types.py
"""

from typing import Any

from pydantic import Field as PydanticField


def Field(
    default: Any = ...,
    description: str = None,
    ge: float = None,
    le: float = None,
    min_length: int = None,
    max_length: int = None,
    regex: str = None,
    choices: list[str | int | float] = None,
) -> Any:
    """Create a field definition with validation rules for Timbal models.

    Args:
        default: Default value for the field. Leave blank for required fields.
        description: Human-readable description of the field.
        ge: Greater than or equal to (minimum value).
        le: Less than or equal to (maximum value).
        min_length: Minimum length for strings/sequences.
        max_length: Maximum length for strings/sequences.
        regex: Regular expression pattern for string validation.
        choices: List of valid values for this field.

    Returns:
        A Pydantic Field instance with the specified validation rules.

    Note:
        - Required fields: Omit default.
        - Optional fields: Set default=None explicitly.
        - The choices parameter extends Pydantic's functionality to match Cog's spec.
    """
    field_info = {
        "description": description,
        "ge": ge,
        "le": le,
        "min_length": min_length,
        "max_length": max_length,
        "pattern": regex,
        "json_schema_extra": {},
    }

    # The `choices` parameter is deprecated in Pydantic v2.
    # Instead, the user should use `Literal[...]` to specify the valid values.
    if choices is not None:
        field_info["json_schema_extra"]["choices"] = choices

    return PydanticField(default, **field_info)



import asyncio
import copy
import inspect
import re
import time
import traceback
from collections import defaultdict
from collections.abc import Callable
from typing import Any

import structlog
from pydantic import BaseModel, Field, TypeAdapter
from uuid_extensions import uuid7

from ..errors import (
    DataKeyError,
    FlowExecutionError,
    InvalidLinkError,
    StepExecutionError,
    StepKeyError,
)
from ..state import RunContext, Snapshot
from ..state.data import (
    BaseData,
    DataError,
    DataMap,
    DataValue,
    get_data_key,
)
from ..state.savers.base import BaseSaver
from ..steps.llms.gateway import handler as llm
from ..types import (
    Message,
    TextContent,
    Tool,
    ToolResultContent,
    ToolUseContent,
)
from ..types.events import (
    FlowOutputEvent,
    StepChunkEvent,
    StepOutputEvent,
    StepStartEvent,
)
from ..types.models import create_model_from_fields, dump, merge_model_fields
from .base import BaseStep
from .link import Link
from .step import Step
from .stream import AsyncGenState, handle_event, sync_to_async_gen
from .utils import Dag, get_ancestors, get_sources, get_successors, is_dag

logger = structlog.get_logger("timbal.graph.flow")

RunnableLike = BaseStep | Callable[..., Any]


class Flow(BaseStep):
    """A class representing a flow of connected steps for executing complex workflows.

    A Flow is a directed acyclic graph (DAG) where nodes are steps and edges define execution order
    and data dependencies. Steps can be functions, LLMs, or other flows. The Flow class handles:

    - Step management and execution ordering
    - Data passing between steps
    - Parameter validation
    - LLM memory management
    - Tool integration for LLM agents
    - State persistence
    - Streaming results

    Attributes:
        steps: Dictionary mapping step IDs to step instances
        links: Dictionary mapping link IDs to link instances
        data: Dictionary storing flow data and mappings
        outputs: Dictionary mapping output names to data keys
        state_saver: Optional saver for persisting flow state
        is_llm: Whether this flow acts as an LLM (always False)
        is_coroutine: Whether this flow returns a coroutine
        is_async_gen: Whether this flow returns an async generator
        _is_compiled: Whether the flow has been compiled
    """
    
    def __init__(
        self, 
        id: str = "flow", 
        path: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a new Flow instance.

        Args:
            id: Optional identifier for the flow. Defaults to "flow".
            path: Optional path for the flow. Use when directly adding this to a flow or subflow.
            **kwargs: Keyword arguments will be passed to the BaseStep's __init__ method.
        """
        if path is None:
            path = id
        super().__init__(id=id, path=path, **kwargs)

        self.steps: dict[str, BaseStep] = {}
        self.links: dict[str, Link] = {}
        self.data: dict[str, BaseData] = {}
        self.outputs: dict[str, str] = {}

        self.state_saver: BaseSaver | None = None

        # ? Think if a subflow could ever act as an LLM (for the moment no, because it does not return anthropic or openai events).
        self.is_llm = False
        # These are used to determine if we need to run the step in an executor.
        self.is_coroutine = False
        self.is_async_gen = True

        self._is_compiled = False

    
    def get_dags(self) -> tuple[Dag, Dag]:
        """Returns the DAG representing the flow's execution order and the reverse DAG for the flow."""
        if hasattr(self, '_dag') and hasattr(self, '_rev_dag'):
            return self._dag, self._rev_dag

        dag = {step_id: set() for step_id in self.steps}
        rev_dag = {step_id: set() for step_id in self.steps}
        for link in self.links.values():
            dag[link.step_id].add(link.next_step_id)
            rev_dag[link.next_step_id].add(link.step_id)
        return dag, rev_dag


    def get_sources(self) -> list[str]:
        """Returns the sources of the flow (i.e. steps with no incoming edges)."""
        if hasattr(self, '_sources'): 
            return self._sources

        dag, _ = self.get_dags()
        sources = get_sources(dag)
        return list(sources)

    
    def params_model(self) -> BaseModel:
        """Returns the Pydantic model defining the expected parameters for this step."""
        if hasattr(self, '_params_model'):
            return self._params_model

        params_model_fields = defaultdict(list)
        for data_key, data_value in self.data.items():
            if not isinstance(data_value, DataMap):
                continue
            map_key = data_value.key
            if map_key in self.data or "." in map_key:
                continue
            step_id, step_param_name = data_key.split(".")
            step = self.steps[step_id]
            step_params = step.params_model()
            step_field_info = step_params.model_fields[step_param_name]
            params_model_fields[map_key].append(step_field_info)

        for param_name, fields_infos in params_model_fields.items():
            if len(fields_infos) == 1:
                params_model_fields[param_name] = fields_infos[0]
            else: # len(steps_fields_infos) > 1
                params_model_fields[param_name] = merge_model_fields(*fields_infos)

        params_model_name = f"Flow_{self.id}_params"
        params_model = create_model_from_fields(
            name=params_model_name,
            model_fields=params_model_fields
        )
        return params_model
        

    def params_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the step's parameter model."""
        if hasattr(self, '_params_model_schema'):
            return self._params_model_schema

        return self.params_model().model_json_schema()


    def return_model(self) -> Any:
        """Returns the expected return type for this step."""
        if hasattr(self, '_return_model'):
            return self._return_model

        fields = {}
        # auto_loops = self.get_auto_loops()
        for return_param_name, data_key in self.outputs.items():

            # If there's a match in the data dictionary, we create a model with the type of the data value.
            try: 
                data_value = get_data_key(self.data, data_key)
                fields[return_param_name] = Field(...)
                fields[return_param_name].annotation = type(data_value)
                continue
            # KeyError is raised when we found a match in the data dictionary, but the specifier 
            # is wrong (e.g. attempting to access a property of a dict that does not exist).
            except KeyError:
                fields[return_param_name] = Field(...)
                fields[return_param_name].annotation = Any
                continue
            except DataKeyError:
                pass

            data_key_parts = data_key.split(".")
            if len(data_key_parts) == 2 and data_key_parts[-1] == "return":
                step_id = data_key_parts[0]
                step = self.steps[step_id]
                step_return_model = step.return_model()
                # Generate a FieldInfo with the model.
                fields[return_param_name] = Field(...)
                fields[return_param_name].annotation = step_return_model
            # TODO If step_id.return.0 -> fetch the type of the list
            # TODO If step_id.return.abc -> fetch the type of the dict value / base model field
            # ? Perhaps we could create an analogous function to get_data_key for navigating the fields definitions
            else:
                fields[return_param_name] = Field(...)
                fields[return_param_name].annotation = Any
        
        if not len(fields):
            return None

        return_model_name = f"Flow_{self.id}_return"
        return_model = create_model_from_fields(
            name=return_model_name, 
            model_fields=fields
        )
        return return_model


    def return_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the flow return model."""
        if hasattr(self, '_return_model_schema'):
            return self._return_model_schema

        return TypeAdapter(self.return_model()).json_schema()


    def _collect_outputs(self, data: dict[str, BaseData]) -> Any:
        """Aux method to collect the outputs from the data dictionary."""

        errors = set()
        outputs = {}
        for output_name, data_key in self.outputs.items():
            # If we don't find the key, we set the output to None.
            # During the flow execution, a step could not be executed, thus its output will not be defined.
            try:
                output_value = get_data_key(data, data_key)
                if isinstance(output_value, DataError): 
                    flow_execution_error_key = f"{self.path}.{data_key.split('.')[0]}"
                    errors.add(flow_execution_error_key)
                else:
                    outputs[output_name] = output_value
            except DataKeyError:
                outputs[output_name] = None
        
        if len(errors):
            raise FlowExecutionError(f"Error collecting outputs {errors}.")
        return outputs
    

    @staticmethod
    def _cut_memory(memory: list[Message], max_window_size: int) -> list[Message]:
        """Cuts the memory list to maintain the specified window size while preserving tool context.

        This method ensures that when truncating the conversation history:
        1. The memory stays within the specified max window size
        2. Tool use/result pairs remain together - if a tool result is in the window, 
           its corresponding tool use call is also preserved
        
        Args:
            memory: List of Message objects representing the conversation history
            max_window_size: Maximum number of messages to keep in the window

        Returns:
            list[Message]: Truncated memory list that preserves tool context
            
        Example:
            If max_window_size=3 but message 4 contains a tool result referencing 
            a tool use in message 1, the returned window will include message 1 
            to maintain the tool context:
            [msg1(tool_use), msg2, msg3, msg4(tool_result)] -> [msg1, msg3, msg4]
        """
        if len(memory) <= max_window_size:
            return memory
        
        final_window_size = max_window_size
        window_tool_use_ids = set()
        window_tool_result_ids = set()
        for message in memory[-max_window_size:]:
            for content in message.content:
                if isinstance(content, ToolResultContent):
                    window_tool_result_ids.add(content.id)
                if isinstance(content, ToolUseContent):
                    window_tool_use_ids.add(content.id)
        window_tool_result_ids_missing = window_tool_result_ids - window_tool_use_ids

        i = len(memory) - max_window_size - 1
        while i >= 0:
            if not len(window_tool_result_ids_missing):
                break
            message = memory[i]
            for content in message.content:
                if isinstance(content, ToolUseContent):
                    if content.id in window_tool_result_ids_missing:
                        window_tool_result_ids_missing.remove(content.id)
                        # ? We could study grabbing the message that triggered the tool call as well for context. 

            final_window_size += 1
            i -= 1

        return memory[-final_window_size:]


    def _resolve_step_args(
        self, 
        step_id: str, 
        data: dict[str, BaseData],
    ) -> dict[str, Any]:
        """Resolves the arguments for a step given a data state.

        Args:
            step_id: Identifier of the step to resolve arguments for
            data: Dictionary containing all flow data, including inputs and previous step results

        Returns:
            Dictionary containing the resolved arguments for the step

        Note:
            These arguments won't be validated against the step's parameter model. 
            Validation must be performed by the caller.
        """
        step = self.steps[step_id]
        step_params_model = step.params_model()
        step_args = {}

        for step_param_name in step_params_model.model_fields.keys():
            step_param_key = f"{step_id}.{step_param_name}"
            step_param_data = data.get(step_param_key, None)

            if step_param_data is None:
                continue

            step_param = step_param_data.resolve(context_data=data)

            # This param will go directly to the step params model validation.
            # We don't want to pass an explicit None. 
            # The pydantic model validation will handle this if the field is optional.
            if step_param is None:
                continue

            step_args[step_param_name] = step_param
        
        _, rev_dag = self.get_dags()

        for ancestor_id in rev_dag[step_id]:
            link = self.links[f"{ancestor_id}-{step_id}"]

            if link.is_tool:
                ancestor_output = data.get(f"{link.step_id}.return", None)
                if not ancestor_output:
                    continue
                ancestor_output = ancestor_output.resolve(context_data=data)
                if not isinstance(ancestor_output, Message):
                    continue
                for content in ancestor_output.content:
                    if isinstance(content, ToolUseContent):
                        if content.name == step_id:
                            step_args = {**content.input, **step_args}

            if link.is_tool_result:
                ancestor_output = data.get(f"{link.step_id}.return", None)
                if ancestor_output is None:
                    continue 
                ancestor_output = ancestor_output.resolve(context_data=data)
                # Check memory to fetch the tool use ID. SDKs will need this to be able to match results to tool calls.
                step_memory = data.get(f"{step_id}.memory", None)
                if step_memory is None:
                    continue
                step_memory = step_memory.resolve(context_data=data)
                step_memory_tool_result_inserted = False
                for message in step_memory[::-1]: # Check in reverse. Tool calls will likely be the last messages.
                    for content in message.content:
                        if isinstance(content, ToolUseContent):
                            if content.name == link.step_id and not step_memory_tool_result_inserted:
                                tool_result_message = Message(
                                    role="user",
                                    content=[ToolResultContent(
                                        id=content.id,
                                        content=[TextContent(text=str(ancestor_output))] # TODO Rethink str() cast.
                                    )]
                                )
                                step_memory.append(tool_result_message)
                                step_memory_tool_result_inserted = True

        # Hack for injecting a prompt into the LLM's memory as last message.
        if step.is_llm:
            step_prompt = step_args.pop("prompt", None)
            if step_prompt is not None:
                # Ensure the prompt is a user message.
                if isinstance(step_prompt, Message) and step_prompt.role != "user":
                    step_prompt = Message(
                        role="user",
                        content=step_prompt.content
                    )
                # If the prompt comes from an unvalidated pydantic model field, we need to validate it.
                else:
                    step_prompt = Message.validate(step_prompt)
                
                if not isinstance(step_prompt, Message):
                    raise ValueError(f"Prompt must be an instance of Message. Got {type(step_prompt)} for step {f'{self.path}.{step_id}'}.")
                
                if "memory" in step_args:
                    step_args["memory"].append(step_prompt)
                    # Limit the memory size. This is individual per llm step. Hence, we can have multiple LLMs pointing
                    # to the same memory key but have different memory window sizes.
                    if step._memory_window_size is not None:
                        step_args["memory"] = self._cut_memory(step_args["memory"], step._memory_window_size)
                else:
                    step_args["memory"] = [step_prompt]

        return step_args


    async def _run_step(
        self, 
        step_id: str, 
        data: dict[str, Any],
        context: RunContext,
    ) -> tuple[Any, Any]:
        """Executes a single step in the flow and returns its result.

        Resolves step arguments from the data dictionary, validates them against the step's parameter model,
        and executes the step. Handles both synchronous and asynchronous step execution.

        Args:
            step_id: Identifier of the step to execute
            data: Dictionary containing all flow data, including inputs and previous step results
            context: Run context


        Returns:
            The step's input arguments
            The step's output (direct value, generator, or awaited coroutine)

        Note:
            While this function is async, it doesn't await generator results - those are handled by the caller.
            Only direct coroutines from step.run() are awaited here.
        """
        step = self.steps[step_id]

        step_input = self._resolve_step_args(step_id, data)
        # We need to copy to avoid the LLM memories being modified by reference.
        step_input_dump = dump(step_input)

        try:
            # Flow input is validated in the .run() method. That is because we want to validate
            # the first initial call to the parent flow as well.
            if not isinstance(step, Flow):
                step_input = dict(step.params_model().model_validate(step_input))
            # If we're dealing with a regular sync function, we need to run it in an executor to 
            # avoid blocking the event loop.
            if not step.is_coroutine and not step.is_async_gen:
                loop = asyncio.get_running_loop()
                step_result = await loop.run_in_executor(None, lambda: step.run(
                    context=context,
                    **step_input
                ))
                if inspect.isgenerator(step_result):
                    return step_input, sync_to_async_gen(step_result, loop)
                return step_input, step_result
            
            step_result = step.run(
                context=context,
                **step_input
            )

            if step.is_coroutine:
                step_result = await step_result
            
            return step_input_dump, step_result

        except Exception as e:
            raise StepExecutionError(step_input_dump, e) from e


    async def run(
        self, 
        context: RunContext | None = None,
        **kwargs: Any
    ) -> Any:
        """Executes the step's processing logic.
        
        Args:
            context: RunContext
            **kwargs: Additional keyword arguments required for step execution.
        
        Returns:
            Any: The step's processing result. Can be any object, a coroutine or an async generator.
        """
        if context is None:
            context = RunContext(id=uuid7(as_type="str"))
        elif context.id is None:
            context.id = uuid7(as_type="str")

        t0 = int(time.time() * 1000)

        # Validate kwargs to store the validated inputs in the snapshot.
        try:
            kwargs = dict(self.params_model().model_validate(kwargs))
        except Exception as e:
            error = {
                "type": type(e).__name__,
                "message": str(e),
                "traceback": traceback.format_exc(),
            }
            t1 = int(time.time() * 1000)
            if self.state_saver is not None:
                snapshot = Snapshot(
                    v="0.2.0",
                    id=context.id,
                    parent_id=context.parent_id,
                    path=self.path,
                    input=kwargs,
                    output=None,
                    error=error,
                    t0=t0,
                    t1=t1,
                    # We allow for a direct reference to the data dictionary, because we know for 
                    # sure that the PUT operation will not modify the data dictionary in any way.
                    data=self.data,
                )
                # TODO Allow for async put.
                self.state_saver.put(snapshot=snapshot, context=context)
            raise FlowExecutionError(f"Error validating kwargs for step {self.path}.") from e
        

        # Copy the data to avoid potential bugs with data being modified by reference.
        data = copy.deepcopy(self.data)
        data.update({k: DataValue(value=v) for k, v in kwargs.items()})

        # Here we'll store all the steps outputs and run data.
        steps = {}

        # Load LLM memories.
        # Hence if this is a root run (no parent_id), we don't need to load any previous snapshot.
        # TODO Optimize this. There's no need to load previous snapshot if there are no memories to load.
        if self.state_saver is not None and context.parent_id is not None:
            # TODO Allow for async get_last.
            # TODO Think error handling here. What if getting the last snapshot errors?
            last_snapshot = self.state_saver.get_last(path=self.path, context=context)
            if last_snapshot is not None:
                last_snapshot_data = last_snapshot.data
                window_sizes = {}

                for _, step in self.steps.items():
                    if step.is_llm and hasattr(step, "_memory_key"):
                        memory_key = step._memory_key
                        if memory_key not in window_sizes:
                            window_sizes[memory_key] = step._memory_window_size
                        elif window_sizes[memory_key] is not None and step._memory_window_size is not None:
                            window_sizes[memory_key] = max(window_sizes[memory_key], step._memory_window_size)
                        else:
                            window_sizes[memory_key] = None
                        data[memory_key] = copy.deepcopy(last_snapshot_data[memory_key])
                        # We could defer this to pydantic inbuilt validation. We do it here to avoid issues when
                        # processing tools results, where code expects memory to contain a list of proper messages.
                        if isinstance(data[memory_key], DataValue):
                            data[memory_key] = DataValue(value=[Message.validate(message) for message in data[memory_key].resolve()])
                
                # Limit the memory sizes. Take into account the maximum amount of memory required by any of the steps.
                # If we can cut it, we cut it to avoid storing always all the data (potentially growing indefinitely).
                for memory_key, max_window_size in window_sizes.items():
                    if max_window_size is not None:
                        if max_window_size == 0:
                            data[memory_key] = DataValue(value=[])
                        else:
                            memory = data[memory_key].resolve(context_data=last_snapshot_data)
                            if len(memory) > max_window_size:
                                memory = self._cut_memory(memory, max_window_size)
                                data[memory_key] = DataValue(value=memory)

        # Compute (or retrieve the pre-computed) DAGs that determine the flow's execution order.
        dag, rev_dag = self.get_dags()

        # Create events for each step to track their completion.
        tasks_completions = {step_id: asyncio.Event() for step_id in self.steps}

        # When working with agents, we'll need to keep track of the steps that don't need to be executed.
        steps_skipped = set()
        # When an error occurs, we'll need to keep track of the steps that need to be skipped.
        steps_to_skip = set()

        # Create tasks for the flow's sources.
        sources_ids = self.get_sources()
        tasks = []
        start_times = {}
        for source_id in sources_ids:
            yield StepStartEvent(
                run_id=context.id,
                parent_step_id=self.id,
                step_id=source_id,
            )
            task = asyncio.create_task(self._run_step(
                step_id=source_id, 
                data=data,
                context=context,
            ))
            task.step_id = source_id
            tasks.append(task)
            start_times[source_id] = int(time.time() * 1000)

        async_gens: dict[str, AsyncGenState] = {}

        while tasks:
            done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            tasks[:] = pending
            
            for completed_task in done:
                step_id = completed_task.step_id
                step_path = self.steps[step_id].path
                step_type = "flow" if isinstance(self.steps[step_id], Flow) else "step"
                step_input = None
                step_result = None
                step_error = None
                step_usage = None

                if step_id in async_gens:
                    # When dealing with an async generator, we need to handle any possible
                    # exceptions that might occur while consuming the generator.
                    async_gen_state = async_gens[step_id]
                    step_input = async_gen_state.input
                    step_usage = async_gen_state.usage
                    try:
                        step_chunk = await completed_task
                        task = asyncio.create_task(async_gen_state.gen.__anext__())
                        task.step_id = step_id
                        tasks.append(task)
                        step_chunk = handle_event(event=step_chunk, async_gen_state=async_gen_state)
                        if step_chunk is not None:
                            yield StepChunkEvent(
                                run_id=context.id,
                                parent_step_id=self.id,
                                step_id=step_id,
                                step_chunk=step_chunk,
                            )
                        continue

                    except StopAsyncIteration:
                        step_result = async_gen_state.results
                        # Tool results are not marked as llm steps. These can include citations (e.g. perplexity).
                        # We add these citations directly formatting the markdown of the text result.
                        if hasattr(async_gen_state, "citations"):
                            step_citations = async_gen_state.citations
                            if isinstance(step_citations, list) and len(step_citations):
                                for i, citation in enumerate(step_citations, start=1):
                                    step_result[-1]["text"] = step_result[-1]["text"].replace(f"[{i}]", f"[{i}]({citation})")
                        # If the step is an LLM, we add the result message to the corresponding memory.
                        if self.steps[step_id].is_llm:
                            step_result = Message.validate({
                                "role": "assistant",
                                "content": step_result,
                            })
                            # Add the message to the appropriate memory.
                            memory_key = f"{step_id}.memory"
                            if memory_key in data:
                                memory = data[memory_key]
                                # TODO We could handle an arbitrary number of nested memories here
                                if isinstance(memory, DataMap): 
                                    assert memory.key.endswith(".memory"), \
                                        f"Memory data map should point to a memory key. Found '{memory.key}'."
                                    memory_key = memory.key
                                memory = memory.resolve(context_data=data)
                                memory.append(step_result)
                                data[memory_key] = DataValue(value=memory)

                    except Exception as e:
                        step_error = {
                            "type": type(e).__name__,
                            "message": str(e),
                            "traceback": traceback.format_exc(),
                        }
                        # Stop all the successors to this step, without interrupting the execution
                        # of other possible branches in the flow.
                        step_successors = get_successors(step_id, dag)
                        steps_to_skip.update(step_successors)
                        logger.error(
                            f"Step '{step_path}' failed. Skipping successors {step_successors}.",
                            error=step_error,
                        )
                
                else:
                    # Else the try except is handled in the _run_step method.
                    try:
                        step_input, step_result = await completed_task
                        if inspect.isasyncgen(step_result):
                            async_gens[step_id] = AsyncGenState(
                                gen=step_result,
                                input=step_input,
                            )
                            task = asyncio.create_task(step_result.__anext__())
                            task.step_id = step_id
                            tasks.append(task)
                            continue
                
                    except StepExecutionError as e:
                        step_input = e.input
                        step_error = {
                            "type": type(e).__name__,
                            "message": str(e),
                            "traceback": traceback.format_exc(),
                        }
                        # Stop all the successors to this step, without interrupting the execution
                        # of other possible branches in the flow.
                        step_successors = get_successors(step_id, dag)
                        steps_to_skip.update(step_successors)
                        logger.error(
                            f"Step '{step_path}' failed. Skipping successors {step_successors}.",
                            error=step_error,
                        )

                # Store the result in the data dictionary. These will be used for data maps and 
                # collecting outputs from the overall flow execution.
                if step_error is None:
                    data[f"{step_id}.return"] = DataValue(value=step_result)
                else:
                    data[f"{step_id}.return"] = DataError(error=step_error)
                
                # Mark this node as completed. This will unblock the tasks waiting for this node.
                tasks_completions[step_id].set()
                
                # Create tasks for the step successors.
                for successor_id in dag[step_id]:
                    if successor_id in steps_to_skip:
                        continue

                    ancestors = [
                        (ancestor_id, self.links[f"{ancestor_id}-{successor_id}"])
                        for ancestor_id in rev_dag[successor_id]
                    ]

                    if all(
                        tasks_completions[ancestor_id].is_set() 
                        for ancestor_id, _ in ancestors
                        if ancestor_id not in steps_skipped
                    ):
                        # Link conditions are only evaluated once all the ancestors are completed.
                        # If any of the conditions is true, we start the next step.
                        if any(ancestor_link.evaluate_condition(data) for _, ancestor_link in ancestors):
                            yield StepStartEvent(
                                run_id=context.id,
                                parent_step_id=self.id,
                                step_id=successor_id,
                            )
                            task = asyncio.create_task(self._run_step(
                                step_id=successor_id, 
                                data=data,
                                context=context,
                            ))
                            task.step_id = successor_id
                            tasks.append(task)
                            start_times[successor_id] = int(time.time() * 1000)
                        else:
                            steps_skipped.add(successor_id)
                
                start_time = start_times[step_id]
                end_time = int(time.time() * 1000)

                steps[step_path] = {
                    "type": step_type,
                    "input": step_input,
                    "output": step_result,
                    "error": step_error,
                    "t0": start_time,
                    "t1": end_time,
                    "usage": step_usage,
                }

                # TODO This will need to be coherent with the above.
                yield StepOutputEvent(
                    run_id=context.id,
                    parent_step_id=self.id,
                    step_id=step_id,
                    input=step_input,
                    output=step_result,
                    error=step_error,
                    t0=start_time,
                    t1=end_time,
                    usage=step_usage,
                )

        # Grab the properties as defined in the return_model.
        output = None
        error = None
        exception = None
        try: 
            output = self._collect_outputs(data)
        except FlowExecutionError as e:
            error = {
                "type": type(e).__name__,
                "message": str(e),
                "traceback": traceback.format_exc(),
            }
            exception = e

        t1 = int(time.time() * 1000)

        if self.state_saver is not None:
            snapshot = Snapshot(
                v="0.2.0",
                id=context.id,
                parent_id=context.parent_id,
                path=self.path,
                input=kwargs,
                output=output,
                error=error,
                t0=t0,
                t1=t1,
                steps=steps,
                # We allow for a direct reference to the data dictionary, because we know for 
                # sure that the PUT operation will not modify the data dictionary in any way.
                data=data,
            )
            # TODO Allow for async put.
            self.state_saver.put(snapshot=snapshot, context=context)

        if exception is not None:
            raise exception
        
        yield FlowOutputEvent(
            run_id=context.id,
            step_id=self.id,
            output=output,
        )

    
    async def complete(
        self,
        context: RunContext | None = None,
        **kwargs: Any
    ) -> FlowOutputEvent:
        """Flow.run() wrapper method that completes the flow execution.
        
        Args: 
            context: RunContext
            **kwargs: Additional keyword arguments required for step execution.
        
        Returns:
            dict[str, Any]: The flow's selected outputs.
        """
        async for event in self.run(context=context, **kwargs):
            if isinstance(event, FlowOutputEvent):
                return event


    def add_step(
        self,
        step: str | RunnableLike,
        action: RunnableLike | None = None,
        # TODO timeout: int | None = None,
        # TODO retry: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> "Flow":
        """Add a step to the flow.

        Args:
            step: Either:
                - A string ID to identify the step (requires action parameter)
                - The step implementation (BaseStep or callable)
            action: The step implementation when using string ID
            **kwargs: Additional arguments to set as step parameters. Must match the step's param model fields

        Returns:
            The flow instance for chaining

        Raises:
            ValueError: If a step with the same id already exists in the flow
            NotImplementedError: If step type is invalid
            StepKeyError: If any kwarg doesn't match a valid step parameter
        """
        if isinstance(step, str) and action is None:
            raise ValueError("Action is required when adding a step by id.")
        
        if (isinstance(step, BaseStep | Callable) and action is not None and isinstance(action, BaseStep | Callable)):
            raise ValueError("Cannot add a step with two actions.")
        
        if isinstance(step, str):
            id = step
            step = action
        elif callable(step):
            id = step.__name__
        elif isinstance(step, BaseStep):
            # Unfortunatelly we cannot grab the object name since python does not have access 
            # to the variable name the flow is assigned to.
            id = step.id
        
        if id in self.steps:
            raise ValueError(f"Step {id} already exists in the flow.")

        def _update_path(step: BaseStep, path: str) -> None:
            if isinstance(step, Step):
                step.path = f"{path}.{step.id}"
                return
            elif isinstance(step, Flow):
                step.path = f"{path}.{step.id}"
                for sub_step in step.steps.values():
                    _update_path(sub_step, step.path)
            else:
                raise NotImplementedError(f"Invalid step type {type(step)}.")

        if isinstance(step, BaseStep):
            step = copy.deepcopy(step)
            step.id = id
            # This recursively updates recursively the steps paths so that we can uniquely 
            # identify each step, and subflow within the parent flow.
            _update_path(step, self.path)
            # When adding a subflow, ensure it is compiled with all the possible optimizations.
            if isinstance(step, Flow) and not step._is_compiled:
                step.compile()
            self.steps[id] = step
        elif callable(step):
            step = Step(id=id, path=f"{self.path}.{id}", handler_fn=step)
            self.steps[id] = step
        else:
            raise NotImplementedError(f"Invalid step type {step}.")

        if not kwargs:
            return self

        # We use the additional kwargs to set the step's params.
        step_params_model_fields = step.params_model().model_fields
        
        for k, v in kwargs.items():
            if k not in step_params_model_fields:
                raise StepKeyError(f"Parameter {k} does not exist in step {id}.")
            data_key = f"{id}.{k}"

            if isinstance(v, DataMap):
                self.set_data_map(data_key, v.key)
            else:
                # Otherwise create a DataValue
                self.set_data_value(data_key, v)

        return self
        
    
    def add_llm(
        self, 
        id: str | None = None, 
        model: str = "gpt-4o-mini", 
        memory_id: str | None = None,
        memory_window_size: int | None = None,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
        tool_choice: dict[str, Any] | str | None = None,
        frequency_penalty: float | None = None,
        logprobs: bool | None = None,
        top_logprobs: int | None = None,
        presence_penalty: float | None = None,
        seed: int | None = None,
        stop: str | list[str] | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
        parallel_tool_calls: bool | None = None,
        json_schema: dict | None = None,
    ) -> "Flow":
        """Adds an LLM step to the flow.

        This method is just a wrapper around add_step that sets the LLM step parameters,
        and exposes the parameters to the user for linters and IDEs.
        Additionally, it marks the step as an LLM step to implement agentic behavior.

        Args:
            id: Unique identifier for this LLM step
            model: Name of the LLM model to use (e.g. "gpt-4-mini")
            memory_id: Optional memory id to use for this LLM. If None, no memory will be used. 
                       If same as step id, creates new memory. If different, reuses existing memory.
            memory_window_size: Maximum number of messages to keep in the memory.
            max_tokens: Maximum number of tokens to generate.
            system_prompt: System prompt to guide the LLM's behavior and role.
            tool_choice: How the model should use the provided tools.
            frequency_penalty: Positive values penalize token frequency to reduce repetition. 
                               Ranges from -2.0 to 2.0.
            logprobs: Whether to return log probabilities.
            top_logprobs: Number of top log probabilities to return.
            presence_penalty: Positive values penalize tokens based on presence to encourage new topics. 
                              Ranges from -2.0 to 2.0.
            seed: Seed for deterministic sampling.
            stop: Where the model will stop generating.
            temperature: Amount of randomness injected into the response.
            top_p: Nucleus sampling parameter.
            top_k: Only sample from the top K options for each subsequent token.
            parallel_tool_calls: Whether to call tools in parallel.
            json_schema: JSON schema to use for this LLM.

        Returns:
            Flow: The flow instance for method chaining

        These are some of the models that could be used:
        - OpenAI: gpt-4o, gpt-4o-mini, o1, o3-mini, o1-mini
        - Anthropic: claude-3-5-sonnet-20241022, claude-3-5-haiku-20241022, 
        claude-3-opus-20240229, claude-3-sonnet-20240229, claude-3-haiku-20240307
        - TogetherAI: deepseek-ai/DeepSeek-R1, deepseek-ai/DeepSeek-V3, 
        meta-llama/Llama-3.3-70B-Instruct-Turbo, meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo, 
        meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo, meta-llama/Llama-3.2-3B-Instruct-Turbo,
        Qwen/Qwen2.5-Coder-32B-Instruct, Qwen/Qwen2-VL-72B-Instruct, 
        mistralai/Mistral-Small-24B-Instruct-2501, mistralai/Mistral-7B-Instruct-v0.3,
        mistralai/Mixtral-8x22B-Instruct-v0.1, meta-llama/Llama-3.2-11B-Vision-Instruct-Turbo
        - Gemini: gemini-2.0-flash-lite-preview-02-05, gemini-2.0-flash, gemini-1.5-flash, 
        gemini-1.5-flash-8b, text-embedding-004
        """
        add_step_kwargs = {"model": model}

        if system_prompt is not None:
            add_step_kwargs["system_prompt"] = system_prompt
        if max_tokens is not None:
            add_step_kwargs["max_tokens"] = max_tokens
        if tool_choice is not None:
            add_step_kwargs["tool_choice"] = tool_choice
        if frequency_penalty is not None:
            add_step_kwargs["frequency_penalty"] = frequency_penalty
        if logprobs is not None:
            add_step_kwargs["logprobs"] = logprobs
        if top_logprobs is not None:
            add_step_kwargs["top_logprobs"] = top_logprobs
        if presence_penalty is not None:
            add_step_kwargs["presence_penalty"] = presence_penalty
        if seed is not None:
            add_step_kwargs["seed"] = seed
        if stop is not None:
            add_step_kwargs["stop"] = stop
        if temperature is not None:
            add_step_kwargs["temperature"] = temperature
        if top_p is not None:
            add_step_kwargs["top_p"] = top_p
        if top_k is not None:
            add_step_kwargs["top_k"] = top_k
        if parallel_tool_calls is not None:
            add_step_kwargs["parallel_tool_calls"] = parallel_tool_calls
        if json_schema is not None:
            add_step_kwargs["json_schema"] = json_schema
        
        if id is None:
            id = "llm"

        self.add_step(id, llm, **add_step_kwargs)
    
        # Mark the step as an LLM step. This will be used for easier checks in other places to implement agentic behavior.
        self.steps[id].is_llm = True

        if memory_id is not None:
            memory_key = f"{memory_id}.memory"
            # We're reusing the memory of another step.
            if memory_key in self.data:
                data_key = f"{id}.memory"
                self.set_data_map(data_key, memory_key, autolink=False)
            # We're creating a new memory with the specified key. 
            else:
                self.set_data_value(memory_key, [])
                if memory_id != id:
                    self.set_data_map(f"{id}.memory", memory_key, autolink=False)
            
            # Store memory config in the llm step, for resolving the memory at runtime.
            self.steps[id]._memory_key = memory_key
            self.steps[id]._memory_window_size = memory_window_size

        return self


    def remove_step(
        self, 
        id: str,
    ) -> "Flow":
        """Removes a step from the flow.

        This will remove the step from the flow, and all its associated data, outputs and links.
        The expected behavior is that if we were to create a new step with the same id, it would be
        a new step, and not inherit all the old data, outputs and links.

        Args:
            id: The id of the step to remove.

        Returns:
            The flow instance for method chaining.
        """
        links_to_remove = [
            link.id 
            for link in self.links.values()
            if link.step_id == id or link.next_step_id == id
        ]
        for link_id in links_to_remove:
            self.remove_link(link_id)

        data_key_prefix_to_remove = f"{id}."

        outputs_to_remove = [
            output_name 
            for output_name, output_data_key in self.outputs.items() 
            if output_data_key.startswith(data_key_prefix_to_remove)
        ]
        for output_name in outputs_to_remove:
            self.remove_output(output_name)
        
        data_key_prefix_to_remove = f"{id}."
        data_to_remove = [
            data_key 
            for data_key, data in self.data.items() 
            if data_key.startswith(data_key_prefix_to_remove) or
                (isinstance(data, DataMap) and data.key.startswith(data_key_prefix_to_remove))
        ]
        for data_key in data_to_remove:
            self.remove_data(data_key)

        self.steps.pop(id, None)
        return self

    
    # TODO Consider the idea of disabling tool params from here.
    def add_link(
        self, 
        step_id: str, 
        next_step_id: str, 
        condition: str | None = None, 
        is_tool: bool = False,
        description: str | None = None,
        is_tool_result: bool = False,
    ) -> "Flow":
        """Adds a link between two steps.

        Links are used to define the flow execution order. 
        A link from step A to step B means that step B will only start after step A has completed.
        Links can be used so that next steps or previous steps are handled as tools for LLMs.

        Args:
            step_id: The id of the step to link from.
            next_step_id: The id of the step to link to.
            condition: The condition to evaluate to determine if the link should be followed.
                       If the condition is true, the link is followed.
            is_tool: Whether the link is a tool call.
            description: The description of the tool.
            is_tool_result: Whether the link is a tool result.
        
        Returns:
            The flow instance for method chaining.
        """
        if step_id not in self.steps:
            raise InvalidLinkError(f"Step {step_id} not found.")

        if next_step_id not in self.steps:
            raise InvalidLinkError(f"Step {next_step_id} not found.")
        
        link_id = f"{step_id}-{next_step_id}"
        if link_id in self.links:
            raise ValueError(f"Link from {step_id} to {next_step_id} already exists.")

        dag, rev_dag = self.get_dags()
        next_step_ancestors = get_ancestors(node_id=next_step_id, rev_dag=rev_dag)
        if step_id in next_step_ancestors:
            raise InvalidLinkError(f"Step {step_id} is already an ancestor of step {next_step_id}.")

        dag[step_id].add(next_step_id)
        if not is_dag(graph=dag):
            raise InvalidLinkError(f"Link from {step_id} to {next_step_id} creates a cycle.")

        if is_tool:
            if not self.steps[step_id].is_llm:
                raise InvalidLinkError(f"Cannot add link '{step_id}-{next_step_id}' as a tool, " \
                                       f"because step '{step_id}' is not an LLM.")

            tool_input_schema = self.steps[next_step_id].params_model_schema()

            tool = Tool(
                name=next_step_id,
                description=description if description else "",
                input_schema=tool_input_schema,
            )

            existing_tools = self.data.get(f"{step_id}.tools", None)
            if existing_tools:
                assert isinstance(existing_tools, DataValue), "Tools should always be instances of DataValue."
                existing_tools = existing_tools.resolve(context_data=None) 
                existing_tools_names = [existing_tool.name for existing_tool in existing_tools]
            else:
                existing_tools = []
                existing_tools_names = []

            if next_step_id not in existing_tools_names:
                existing_tools.append(tool)
            self.set_data_value(f"{step_id}.tools", existing_tools)
        
        if is_tool_result:
            if not self.steps[next_step_id].is_llm:
                raise InvalidLinkError(f"Cannot add link '{step_id}-{next_step_id}' as a tool result, " \
                                       f"because step '{next_step_id}' is not an LLM.")
            # TODO We could further validate that the next step shares de memory with the previous tool step.
            # ! I don't love this. 'memory' is not an argument of the tool per se. We're abusing the data maps.
            self.set_data_map(f"{step_id}.memory", f"{next_step_id}.memory", autolink=False)
        
        link = Link(
            step_id=step_id, 
            next_step_id=next_step_id, 
            condition=condition, 
            is_tool=is_tool,
            is_tool_result=is_tool_result,
        )
        self.links[link.id] = link

        return self
    

    def remove_link(self, id: str) -> "Flow":
        """Removes a link from the flow.

        Args:
            id: The id of the link to remove.

        Returns:
            The flow instance for method chaining.
        """
        self.links.pop(id, None)
        return self


    def set_input(
        self,
        data_key: str, 
        input_key: str | None = None,
    ) -> "Flow":
        """Sets an input mapping to the flow.

        Args:
            data_key: The key in the data dict to retrieve as an input.
            input_key: Whether to map this input value to a specific key.
        
        Returns:
            The flow instance for method chaining.
        """
        self.set_data_map(data_key, input_key)
        return self
    

    def set_output(
        self, 
        data_key: str,
        output_key: str | None = None, # TODO
    ) -> "Flow":
        """Adds an output mapping to the flow.

        Careful this does not validate the data key or anything.
        If there's a mapping error or missing key it will return an error at runtime.

        This overrides the previous output mapping for the same name.

        Args:
            data_key: The key in the data dict to retrieve as an output.
            output_key: Whether to map this output value to a specific key.
        
        Returns:
            The flow instance for method chaining.
        """
        self.outputs[output_key] = data_key
        return self


    def remove_output(
        self, 
        output_key: str,
    ) -> "Flow":
        """Removes an output mapping from the flow.

        Args:
            output_key: The key of the output to remove.

        Returns:
            The flow instance for method chaining.
        """
        self.outputs.pop(output_key, None)
        return self


    def set_data_value(
        self,
        data_key: str,
        data_value: Any,
        autolink: bool = True,
    ) -> "Flow":
        """Sets a data value for a data key.

        This method does not check if the data key exists in the flow data.
        It also does not check if the data value is valid (e.g. if the data key is a set param).
        Careful, if the data value provided to a step param is not valid, it will raise an error at runtime.

        Args:
            data_key: The data key to set the value for.
            data_value: The value to set.
            autolink: If string interpolation is used to map to other steps values, 
                      whether to automatically add a link between the source and target steps.
        Returns:
            The flow instance for method chaining.
        """
        self.data[data_key] = DataValue(value=data_value)

        if not autolink:
            return self

        if isinstance(data_value, str) and "{{" in data_value and "}}" in data_value:
            target_step_id = data_key.split(".")[0] if "." in data_key else None
            template_refs = re.findall(r"\{\{([\w\.]+)\}\}", data_value)
            
            for ref in template_refs:
                if "." not in ref:
                    continue
                source_step_id = ref.split(".")[0]
                if f"{source_step_id}-{target_step_id}" not in self.links:
                    self.add_link(source_step_id, target_step_id)

        return self

                
    def set_data_map(
        self,
        data_key: str,
        map_key: str,
        default: Any | None = "__NO_DEFAULT__",
        autolink: bool = True,
    ) -> "Flow":
        """Sets a map from a step param to a data key.

        This method does not check if the data key or the map key exist
        in the flow data. This allows for dynamic mapping during runtime (and other cool stuff).
        Careful, if the mapping is not valid at runtime, it will raise an error.

        Args:
            data_key: The data key of the step param to map.
            map_key: The key to map the data key to.
            default: The default value to return if the data key is not found.
            autolink: Whether to automatically add a link between the source and target steps.
        
        Returns:
            The flow instance for method chaining.
        """
        # TODO We could validate that the data key is of an appropriate type.

        if data_key in self.data:
            existing_data = self.data[data_key]
            if isinstance(existing_data, DataMap):
                # ? We could study overriding the existing map.
                raise ValueError(f"Cannot set multiple data maps for the same data key: {data_key}.")
            elif isinstance(existing_data, DataValue):
                # We use the previous data value as the default for the new map.
                if default is None:
                    default = existing_data.value
            else:
                raise NotImplementedError(
                    f"Cannot set data map for data key: {data_key} with value of type {type(existing_data)}.")

        self.data[data_key] = DataMap(key=map_key, default=default)

        if autolink and "." in data_key and "." in map_key:
            target_step_id = data_key.split(".")[0]
            if target_step_id not in self.steps:
                logger.warning(f"Cannot auto-add link for {data_key} -> {map_key}")
                return self

            source_step_id = map_key.split(".")[0]
            if source_step_id not in self.steps:
                logger.warning(f"Cannot auto-add link for {data_key} -> {map_key}")
                return self

            if f"{source_step_id}-{target_step_id}" not in self.links:
                self.add_link(source_step_id, target_step_id)

        return self

    
    def remove_data(
        self,
        key: str,
    ) -> "Flow":
        """Removes a data key from the flow.

        Args:
            key: The data key to remove.

        Returns:
            The flow instance for method chaining.
        """
        self.data.pop(key, None)
        return self


    def compile(
        self,
        state_saver: BaseSaver | None = None,
    ) -> "Flow":
        """Compiles the flow by pre-computing commonly used properties and 'caching' them.

        This method:
        - Pre-computes and caches the DAG and reverse DAG for execution order.
        - Pre-computes and caches the parameter and return models and schemas.
        - Sets the state saver for persisting flow state.

        Args:
            state_saver: Optional BaseSaver instance to persist flow state during execution.
        
        Returns:
            The flow instance for method chaining.
        """
        dag, rev_dag = self.get_dags()
        self._dag = dag
        self._rev_dag = rev_dag

        sources = self.get_sources()
        self._sources = sources

        self._params_model = self.params_model()
        self._params_model_schema = self.params_model_schema()
        self._return_model = self.return_model()
        self._return_model_schema = self.return_model_schema()

        self.state_saver = state_saver

        # Mark the flow as compiled to prevent re-compiling when importing as a subflow.
        self._is_compiled = True

        return self
    


from abc import ABC, abstractmethod

from ..context import RunContext
from ..snapshot import Snapshot


class BaseSaver(ABC):
    """Base class for creating a Flow state saver.

    State savers persist Flow states across multiple interactions.

    Note:
        When creating a custom state saver, consider implementing async
        versions to avoid blocking the main thread.
    """


    @abstractmethod 
    def get_last(
        self, 
        path: str,
        context: RunContext,
    ) -> Snapshot | None:
        """Retrieve the last snapshot matching the specified criteria.

        Args:
            path: Flows are nested structures. We don't ensure id uniqueness across nested subflows.
                  This path ensures uniqueness of every different BaseStep in the flow.
            context: The run context. For instance, if we're looking for a specific parent id or a group id.

        Returns:
            The last snapshot matching the specified criteria.
        """
        pass


    @abstractmethod
    def put(
        self, 
        snapshot: Snapshot, 
        context: RunContext,
    ) -> None:
        """Store a new snapshot.

        Args:
            snapshot: The Snapshot object to store.
            context: The run context. Might be used for authentication to an external service.
        """
        pass



import os
from typing import Any, Literal

from dotenv import load_dotenv
from openai import AsyncOpenAI

from ...types import Field, Message, Tool

load_dotenv()

async def handler(
    # This is not actually used inside the function. It's a hack to have it as an argument
    # so we can use it when mapping data in the flow.
    prompt: Message = Field(default=None, description="Message to send to the LLM."), # noqa: ARG001
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    memory: list[Message] = Field(description="A list containing the conversation history between the user and the LLM"),
    model: Literal["gpt-4o", "gpt-4o-mini", "o1", "o3-mini", "o1-mini"] | str = Field(default="gpt-4o", description="Name of the LLM model to use"),
    tools: list[Tool | dict] = Field(default=None, description="List of tools/functions that the LLM can call"),
    tool_choice: dict[str, Any] | str = Field(default = {"type": "auto"}, description="How the model should use the provided tools"),
    max_tokens: int = Field(default=None, description="The maximum number of tokens in the response"),
    frequency_penalty: float = Field(default=0, description="Penalizes token repetition on a scale of -2.0 to 2.0. Higher values reduce repetition"),
    logprobs: bool = Field(default=False, description="Whether to return logprobs with the returned text"),
    top_logprobs: int = Field(default=None, description="Number of most likely tokens (0-20) to return with probabilities. Requires logprobs=True"),
    presence_penalty: float = Field(default=0, description="Penalizes token presence on a scale of -2.0 to 2.0. Higher values encourage new topics"),
    seed: int = Field(default=None, description="Seed for deterministic sampling (Beta feature)"),
    stop: str | list[str] = Field(default=None, description="Up to 4 sequences where the model will stop generating further tokens"),
    temperature: float = Field(default=1, description="Amount of randomness injected into the response. Ranges from 0.0 to 2.0"),
    top_p: float = Field(default=1, description="Nucleus sampling parameter. Ranges from 0.0 to 1.0"),
    parallel_tool_calls: bool = Field(default=True, description="Whether to enable parallel function calling during tool use"),
    json_schema: dict = Field(default=None, description="The JSON input schema that the model must output"),
) -> Any:
    """This handler manages interactions with OpenAI's API, processing requests with 
    appropriate parameters and returning streaming responses.

    Args:
        prompt: Hack to pass an input to the LLM.
        memory: Chat history containing user and LLM messages.
        system_prompt: Instructions for the LLM to follow.
        model: Name of the OpenAI model to use.
        tools: List of tools/functions the LLM can call.
        tool_choice: How the model should use the provided tools.
        max_tokens: Maximum number of tokens in the response. Defaults to 1024.
        frequency_penalty: Value between -2.0 and 2.0 to penalize token frequency. Defaults to 0.
        logprobs: Whether to return log probabilities. Defaults to False.
        top_logprobs: Number of most likely tokens to return (0-20). Requires logprobs=True.
        presence_penalty: Value between -2.0 and 2.0 to penalize token presence. Defaults to 0.
        seed: Seed for deterministic sampling (Beta feature).
        stop: Sequences where token generation should stop.
        temperature: Sampling temperature between 0 and 2. Defaults to 1.
        top_p: Nucleus sampling probability threshold. Defaults to 1.
        parallel_tool_calls: Whether to execute tool calls in parallel. Defaults to True.
        json_schema: JSON schema for structured output (Anthropic models only).

    Returns:
        Any: The streaming response from the OpenAI API.
    """
    client = AsyncOpenAI(api_key=os.getenv("OPENAI_API_KEY"))
    
    # Get default values for parameters
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    model = model.default if hasattr(model, 'default') else model
    tools = tools.default if hasattr(tools, 'default') else tools
    tool_choice = tool_choice.default if hasattr(tool_choice, 'default') else tool_choice
    max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    frequency_penalty = frequency_penalty.default if hasattr(frequency_penalty, 'default') else frequency_penalty
    logprobs = logprobs.default if hasattr(logprobs, 'default') else logprobs
    top_logprobs = top_logprobs.default if hasattr(top_logprobs, 'default') else top_logprobs
    presence_penalty = presence_penalty.default if hasattr(presence_penalty, 'default') else presence_penalty
    seed = seed.default if hasattr(seed, 'default') else seed
    stop = stop.default if hasattr(stop, 'default') else stop
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    parallel_tool_calls = parallel_tool_calls.default if hasattr(parallel_tool_calls, 'default') else parallel_tool_calls
    json_schema = json_schema.default if hasattr(json_schema, 'default') else json_schema

    messages = []
    if system_prompt:
        messages.append({"role": "system", "content": system_prompt})
    for msg in memory:
        openai_msg = msg.to_openai_input()
        messages.append(openai_msg)

    kwargs = {
        "model": model,
        "messages": messages,
        "stream": True,
        "presence_penalty": presence_penalty,
        "temperature": temperature,
        "top_p": top_p,
        "stream_options": {"include_usage": True},
    }

    # Optional parameters
    if stop:
        kwargs["stop"] = stop
    if frequency_penalty:
        kwargs["frequency_penalty"] = frequency_penalty
    if seed:
        kwargs["seed"] = seed
    if logprobs:
        kwargs["logprobs"] = logprobs
    if top_logprobs:
        kwargs["top_logprobs"] = top_logprobs
    if max_tokens:
        kwargs["max_tokens"] = max_tokens
    
    # Format parameters to match OpenAI's API requirements
    if tools:
        tools_openai = []
        for tool in tools:
            if isinstance(tool, Tool):
                tools_openai.append(tool.to_openai())
            else:
                tools_openai.append(tool)
        kwargs["tools"] = tools_openai
        if isinstance(tool_choice, dict):
            if tool_choice["type"] != "function":
                tool_choice = tool_choice["type"]
        kwargs["tool_choice"] = tool_choice
        kwargs["parallel_tool_calls"] = parallel_tool_calls

    if json_schema:
        kwargs["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "name": "print_json",
                "schema": {
                    "type": "object",
                    "properties": json_schema
                }
            }
        } 

    response = await client.chat.completions.create(**kwargs)

    return response


from typing import Any

import structlog
from pydantic import BaseModel, ConfigDict, computed_field

from ..state.data import DATA_KEY_INTERPOLATION_PATTERN, BaseData, get_data_key
from ..types import Message, ToolUseContent

logger = structlog.get_logger("timbal.graph.link")


class Link(BaseModel):
    """A Link represents a connection between two steps in a workflow graph.
    
    Links can be conditional (using a condition string that evaluates against workflow data),
    or represent tool calls and their results. Each link connects a source step (step_id)
    to a destination step (next_step_id).
    """
    # Allow storing extra fields in the model.
    # Validate dynamic assignments to the model.
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    step_id: str 
    """The ID of the source step."""
    next_step_id: str
    """The ID of the destination step."""
    condition: str | None = None
    """Optional string containing a condition to evaluate."""
    is_tool: bool = False
    """ Whether this link represents a tool call."""
    is_tool_result: bool = False
    """Whether this link represents a tool result."""
    metadata: dict[str, Any] = {}
    """Additional metadata associated with the link."""


    @computed_field
    @property
    def id(self) -> str:
        """The id of the link is the concatenation of the step_id and next_step_id."""
        return f"{self.step_id}-{self.next_step_id}"

    
    def evaluate_condition(self, data: dict[str, Any]) -> bool:
        """Evaluates if this link's condition is met given the current data."""

        if self.is_tool:
            step_output = data.get(f"{self.step_id}.return", None)

            assert isinstance(step_output, BaseData), "Step output should always be an instance of BaseData."
            step_output = step_output.resolve(context_data=data)
            
            # Check if this is an LLM output with tool calls
            if isinstance(step_output, Message):
                for content in step_output.content:
                    if isinstance(content, ToolUseContent) and content.name == self.next_step_id:
                        return True
            return False

        if self.condition:
            # Search for all keys we may have in the context data.
            template_keys = DATA_KEY_INTERPOLATION_PATTERN.findall(self.condition)
            condition = self.condition  # Work with a copy instead of modifying self.condition 
        
            for template_key in template_keys:
                replacement_value = get_data_key(data, template_key)
                # TODO Handle more scenarios here.
                if isinstance(replacement_value, str):
                    replacement_value = f"'{replacement_value}'"
                else:
                    replacement_value = str(replacement_value)
                # In order to evaluate the condition, we need to place the value in the condition as a string.
                condition = condition.replace(f"{{{{{template_key}}}}}", replacement_value)

            if not eval(condition):
                return False
        
        # By default, if the link was not added as a tool, or no additional condition is passed, we always return True.
        return True



import os
from typing import Any, Literal

import structlog
from dotenv import load_dotenv
from openai import AsyncOpenAI

from ...types import Field

load_dotenv()

logger = structlog.get_logger("timbal.steps.perplexity.search")

async def search(
    query: str = Field(description="Query to search for."),
    model: str = Field(default="sonar", description="Model to use."),
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    # max_tokens: int | None = Field(default= None,description="Maximum number of tokens to generate."),
    temperature: float = Field(default=0.2, description="Amount of randomness in the response. Ranges from 0 to 2."),
    top_p: float = Field(default=0.9, description="Nucleus sampling threshold. Ranges from 0 to 1."),
    search_domain_filter: Any = Field(default=None, description="Given a list of domains, limit the citations used by the online model to URLs from the specified domains. Currently limited to only 3 domains for whitelisting and blacklisting. For blacklisting add a - to the beginning of the domain string. Only available in certain tiers"),
    return_images: bool = Field(default=False, description="Whether to return images in the response. Only available in certain tiers."),
    return_related_questions: bool = Field(default=False, description="Whether to return related questions in the response. Only available in certain tiers."),
    search_recency_filter: Literal['month', 'week', 'day', 'hour'] = Field(default=None, description="Returns search results within the specified time interval - does not apply to images."),
    top_k: int = Field(default=0, description="The number of tokens to keep for highest top-k filtering. Ranges from 0 to 2048. If set to 0, top-k filtering is disabled. We recommend either altering top_k or top_p, but not both."),
    stream: bool = Field(default=False, description="Whether or not to incrementally stream the response with server-sent events with content-type: text/event-stream."),
    presence_penalty: float = Field(default=0, description="Presence penalty to use. Ranges from -2.0 to 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. Incompatible with frequency_penalty."),
    frequency_penalty: float = Field(default=1, description="A multiplicative penalty greater than 0. Values greater than 1.0 penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. Incompatible with presence_penalty."),
    response_format: dict = Field(default=None, description="Enable structured outputs with a JSON or Regex schema. Only available in certain tiers. e.g. { type: 'json_schema', json_schema: {'schema': object} }, { type: 'regex', regex: {'regex': str} }"),
) -> Any:
    """This handler manages interactions with Perplexity's API, processing requests with 
    appropriate parameters and returning streaming responses.

    Args:
        query: Query to search for.
        model: Model to use. Defaults to "sonar".
        system_prompt: System prompt to guide the LLM's behavior and role.
        max_tokens: Maximum number of tokens to generate.
        temperature: Amount of randomness in the response. Ranges from 0 to 2.
        top_p: Nucleus sampling threshold. Ranges from 0 to 1.
        search_domain_filter: Given a list of domains, limit the citations used by the online model to URLs from the specified domains. Currently limited to only 3 domains for whitelisting and blacklisting. For blacklisting add a - to the beginning of the domain string. Only available in certain tiers
        return_images: Whether to return images in the response. Only available in certain tiers.
        return_related_questions: Whether to return related questions in the response. Only available in certain tiers.
        search_recency_filter: Returns search results within the specified time interval - does not apply to images.
        top_k: The number of tokens to keep for highest top-k filtering. Ranges from 0 to 2048. If set to 0, top-k filtering is disabled. We recommend either altering top_k or top_p, but not both.
        stream: Whether or not to incrementally stream the response with server-sent events with content-type: text/event-stream.
        presence_penalty: Presence penalty to use. Ranges from -2.0 to 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. Incompatible with frequency_penalty.
        frequency_penalty: A multiplicative penalty greater than 0. Values greater than 1.0 penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. Incompatible with presence_penalty.
        response_format: Enable structured outputs with a JSON or Regex schema. Only available in certain tiers. e.g. { type: 'json_schema', json_schema: {'schema': object} }, { type: 'regex', regex: {'regex': str} }

    Returns:
        Any: The streaming response from the Perplexity API.
    """
    client = AsyncOpenAI(
        api_key=os.getenv('PERPLEXITY_API_KEY'), 
        base_url="https://api.perplexity.ai",
    )

    # Get the actual values from Field objects
    model = model.default if hasattr(model, 'default') else model
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    # max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    search_domain_filter = search_domain_filter.default if hasattr(search_domain_filter, 'default') else search_domain_filter
    return_images = return_images.default if hasattr(return_images, 'default') else return_images
    return_related_questions = return_related_questions.default if hasattr(return_related_questions, 'default') else return_related_questions
    search_recency_filter = search_recency_filter.default if hasattr(search_recency_filter, 'default') else search_recency_filter
    top_k = top_k.default if hasattr(top_k, 'default') else top_k
    stream = stream.default if hasattr(stream, 'default') else stream
    presence_penalty = presence_penalty.default if hasattr(presence_penalty, 'default') else presence_penalty
    frequency_penalty = frequency_penalty.default if hasattr(frequency_penalty, 'default') else frequency_penalty
    response_format = response_format.default if hasattr(response_format, 'default') else response_format

    messages = []
    if system_prompt:
        messages.append({"role": "system", "content": system_prompt})
    messages.append({"role": "user", "content": query})
    
    response = await client.chat.completions.create(
            model=model,
            messages=messages,
            # max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            stream=True,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            response_format=response_format,
        )

    async for chunk in response:
        yield chunk



import inspect
import json

import anthropic
import pytest
from anthropic.types import (
    InputJSONDelta,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    TextDelta,
    ToolUseBlock,
)
from timbal.steps.llms.gateway import handler
from timbal.types import Content, File, Message
from timbal.types.models import create_model_from_argspec


@pytest.mark.asyncio
async def test_basic_response():
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is 2+2?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "claude-3-haiku-20240307"
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, RawContentBlockDeltaEvent):
            if isinstance(chunk.delta, TextDelta):
                response_chunks += chunk.delta.text    
    
    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "4" in response_chunks, "Response should contain the number 4"


@pytest.mark.asyncio
async def test_image_upload():
    url = "https://content.timbal.ai/assets/Three-Australian-Shepherd-puppies-sitting-in-a-field.jpg"
    file = File.validate(url)
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is in this image?"}, {"type": "file", "file": file}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "claude-3-haiku-20240307"
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, RawContentBlockDeltaEvent):
            if isinstance(chunk.delta, TextDelta):
                response_chunks += chunk.delta.text    

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "dog" in response_chunks.lower() or "pup" in response_chunks.lower(), \
        "Response should mention that there are dogs or puppies in the image"


@pytest.mark.asyncio
async def test_tools():
    tools = [
        {
            "name": "get_weather",
            "description": "Get the weather for a location",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The location to get weather for"
                    }
                },
                "required": ["location"]
            }
        }
    ]

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "What's the weather in London?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "claude-3-haiku-20240307",
        "tools": tools,
        "tool_choice": "auto"
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = []
    async for chunk in handler(**dict(model_args)):
        # Start of a tool use or text message
        if isinstance(chunk, RawContentBlockStartEvent):
            if isinstance(chunk.content_block, ToolUseBlock):
                response_chunks.append({
                    "type": "tool_use",
                    "id": chunk.content_block.id,
                    "name": chunk.content_block.name,
                    "input": ""
                })
            else:
                response_chunks.append({
                    "type": "text",
                    "text": chunk.content_block.text
                })
        # Add content to the last message
        elif isinstance(chunk, RawContentBlockDeltaEvent):
            if isinstance(chunk.delta, InputJSONDelta):
                response_chunks[-1]["input"] += chunk.delta.partial_json
            elif isinstance(chunk.delta, TextDelta):
                response_chunks[-1]["text"] += chunk.delta.text
        # Finish a tool use or text message
        elif isinstance(chunk, RawContentBlockStopEvent):
            last_message = response_chunks[-1]
            if last_message["type"] == "tool_use":
                last_message["input"] = json.loads(last_message["input"])
            last_message = Content.model_validate(last_message)
            response_chunks[-1] = last_message

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, list), "Response should be a list"
    weather_tool = next(
        (msg for msg in response_chunks if msg.type == "tool_use" and msg.name == "get_weather"),
        None
    )
    assert weather_tool is not None, "Weather tool call not found"
    assert isinstance(weather_tool.id, str) and weather_tool.id, "Tool call should have a non-empty ID"
    assert isinstance(weather_tool.input, dict), "Tool input should be a dictionary"
    assert weather_tool.input.get("location") == "London", "Tool input should contain 'London'"


@pytest.mark.asyncio
async def test_multiple_tools():
    tools=[
        {
            "name": "get_weather",
            "description": "Get the current weather in a given location",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA"
                    },
                    "unit": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The unit of temperature, either 'celsius' or 'fahrenheit'"
                    }
                },
                "required": ["location"]
            }
        },
        {
            "name": "get_time",
            "description": "Get the current time in a given time zone",
            "input_schema": {
                "type": "object",
                "properties": {
                    "timezone": {
                        "type": "string",
                        "description": "The IANA time zone name, e.g. America/Los_Angeles"
                    }
                },
                "required": ["timezone"]
            }
        }
    ]

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "What is the weather like right now in New York? Also what time is it there?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant. When a user asks multiple questions that require different tools, make sure to use all necessary tools to provide complete information.",
        "model": "claude-3-5-sonnet-20241022",
        "tools": tools,
        "tool_choice": {"type":"auto"},
    }
    model_args = model.model_validate(model_args)

    response_chunks = []
    async for chunk in handler(**dict(model_args)):
        # Start of a tool use or text message
        if isinstance(chunk, RawContentBlockStartEvent):
            if isinstance(chunk.content_block, ToolUseBlock):
                response_chunks.append({
                    "type": "tool_use",
                    "id": chunk.content_block.id,
                    "name": chunk.content_block.name,
                    "input": ""
                })
            else:
                response_chunks.append({
                    "type": "text",
                    "text": chunk.content_block.text
                })

        # Add content to the last message
        elif isinstance(chunk, RawContentBlockDeltaEvent):
            if isinstance(chunk.delta, InputJSONDelta):
                response_chunks[-1]["input"] += chunk.delta.partial_json
            elif isinstance(chunk.delta, TextDelta):
                response_chunks[-1]["text"] += chunk.delta.text
        
        # Finish a tool use or text message
        elif isinstance(chunk, RawContentBlockStopEvent):
            last_message = response_chunks[-1]
            if last_message["type"] == "tool_use":
                last_message["input"] = json.loads(last_message["input"])
            last_message = Content.model_validate(last_message)
            response_chunks[-1] = last_message

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, list), "Response should be a list"
    assert any(
        msg.type == "tool_use" and msg.name == "get_weather" 
        for msg in response_chunks
    ), "Weather tool call not found"
    assert any(
        msg.type == "tool_use" and msg.name == "get_time" 
        for msg in response_chunks
    ), "Time tool call not found"


@pytest.mark.asyncio
async def test_response_format():
    json_schema = {
        "title": {"type": "string"},
        "author": {"type": "string"},
        "genre": {"type": "string"},
        "publication_year": {"type": "integer"},
        "summary": {"type": "string"}
    }  

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "Summarize The Eras Tour by Taylor Swift with fields for title, author, genre, publication year, and a short summary."}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "claude-3-haiku-20240307",
        "json_schema": json_schema
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, RawContentBlockDeltaEvent):
            if isinstance(chunk.delta, InputJSONDelta):
                response_chunks += chunk.delta.partial_json

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert json.loads(response_chunks), "Response should match the JSON schema"


@pytest.mark.asyncio
async def test_invalid_model():
    memory = Message.validate({
        "role": "user",
        "content": [{"type": "text", "text": "Hello"}]
    })
    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "claude-invalid-model"
    }
    model_args = model.model_validate(model_args)

    with pytest.raises(anthropic.NotFoundError):
        async for _ in handler(**dict(model_args)):
            pass


import pytest
from timbal import Agent
from timbal.state import RunContext
from timbal.state.savers import InMemorySaver


@pytest.mark.asyncio
async def test_in_memory():

    agent_id = "custom_name"

    flow = Agent(
        id=agent_id,
        model="gpt-4o-mini",
    ).compile(state_saver=InMemorySaver())

    flow_output_event = await flow.complete(prompt="My name is David")

    assert len(flow.state_saver.snapshots) == 1
    for i, snapshot in enumerate(flow.state_saver.snapshots):
        agent_memory = snapshot.data[f"{agent_id}.memory"].resolve()
        assert len(agent_memory) == (i + 1) * 2

    flow_output_event = await flow.complete(
        context=RunContext(parent_id=flow_output_event.run_id),
        prompt="What's my name?"
    )

    assert len(flow.state_saver.snapshots) == 2
    for i, snapshot in enumerate(flow.state_saver.snapshots):
        agent_memory = snapshot.data[f"{agent_id}.memory"].resolve()
        assert len(agent_memory) == (i + 1) * 2

    await flow.complete(
        context=RunContext(parent_id=flow_output_event.run_id),
        prompt="And what's yours?"
    )

    assert len(flow.state_saver.snapshots) == 3
    for i, snapshot in enumerate(flow.state_saver.snapshots):
        agent_memory = snapshot.data[f"{agent_id}.memory"].resolve()
        assert len(agent_memory) == (i + 1) * 2



"""
Defines the content types for chat messages in Timbal.

Types:
- TextContent: Plain text messages
- FileContent: File attachments
- ToolUseContent: Request to use a specific tool
- ToolResultContent: Result returned by a tool

All message content must be an instance of one of these types.

Usage:

1. Validating a content:
   >>> text_content = Content.model_validate(text_block)
   >>> tool_use_content = Content.model_validate(tool_use_block)

2. Converting a content to the input format required by OpenAI and Anthropic:
   >>> file_content = FileContent(file=File.validate(data_url))
   >>> file_content.to_openai_input()
   >>> file_content.to_anthropic_input()

"""

import base64
import json
import mimetypes
from ast import literal_eval
from typing import Any, Literal

from anthropic.types import (
    TextBlock as AnthropicTextBlock,
)
from anthropic.types import (
    ToolUseBlock as AnthropicToolUseBlock,
)
from openai.types.chat import (
    ChatCompletionMessageToolCall as OpenAIToolCall,
)
from pydantic import BaseModel
from uuid_extensions import uuid7

from ..file import File


class Content(BaseModel):
    """
    A class representing the content of a chat message.
    """

    type: Literal["text", "file", "tool_use", "tool_result"]

    @classmethod 
    def model_validate(cls, value: Any, *args: Any, **kwargs: Any) -> "Content":
        """Validate and convert input formats into a Content instance."""
        # Don't recurse if we're already dealing with a Content instance
        if isinstance(value, Content):
            return value
        
        # cls will be diferent from Content when we call model_validate on one of the subclasses
        if cls is not Content:
            return super().model_validate(value, *args, **kwargs)
        
        if isinstance(value, str):
            return TextContent(text=value)

        if isinstance(value, File):
            return FileContent(file=value)
        
        if isinstance(value, AnthropicTextBlock):
            return TextContent(text=value.text)

        if isinstance(value, AnthropicToolUseBlock):
            return ToolUseContent(
                id=value.id,
                name=value.name,
                input=value.input,
            )

        if isinstance(value, OpenAIToolCall):
            return ToolUseContent(
                id=value.id,
                name=value.function.name,
                input=literal_eval(value.function.arguments),
            )
        
        # TODO Review
        if isinstance(value, dict):
            content_type = value.get("type", None)

            if content_type == "text":
                return TextContent(text=value.get("text"))
            
            # Anthropic's file content type.
            if content_type == "file":
                return FileContent(file=File.validate(value.get("file")))
            
            # OpenAI's file content type.
            if content_type == "image_url":
                return FileContent(file=File.validate(value.get("image_url")['url']))
            
            if content_type == "input_audio":
                return FileContent(file=File.validate(value.get("input_audio")['data']))

            # Anthropic's tool use content.
            if content_type == "tool_use":
                input_value = value.get("input") 
                if isinstance(input_value, str):
                    if input_value != "":
                        input_value = json.loads(input_value)
                    else:
                        input_value = {}
                return ToolUseContent(
                    id=value.get("id"), 
                    name=value.get("name"), 
                    input=input_value,
                )
            
            # OpenAI's tool use content.
            if content_type == "function":
                return ToolUseContent(
                    id=value.get("id"),  
                    name=value["function"]["name"],
                    input=literal_eval(value["function"]["arguments"]),
                )
            
            # Anthropic's tool result content.
            if content_type == "tool_result":
                tool_result_content = value.get("content", [])
                if not isinstance(tool_result_content, list):
                    tool_result_content = [tool_result_content]
                return ToolResultContent(
                    id=value.get("tool_use_id") or value.get("id"), 
                    content=[cls.model_validate(item) for item in tool_result_content],
                )
        
        raise ValueError(f"Invalid content: {value}")


class FileContent(Content):
    """
    This class represents a file content in a chat message. 
    It also provides methods to convert the file content to the input format required by OpenAI and Anthropic.
    """
    type: Literal["file"] = "file"
    file: File


    def to_openai_input(self) -> dict[str, Any]:
        """Convert the file content to the input format required by OpenAI."""
        # Get mime type. 
        # source schemes: bytes, local_path, data, url
        if self.file.__source_scheme__ == "bytes":
            raise ValueError("Cannot convert bytes-source file to OpenAI message content.")
        elif self.file.__source_scheme__ == "data":
            mime = self.file.__source__.split(";")[0].split(":")[1]
        else:
            mime, _ = mimetypes.guess_type(str(self.file.__source__))
        
        # Ensure the file pointer is at the start of the file if we need to read it.
        current_position = self.file.tell()
        if current_position != 0:
            self.file.seek(0)

        if mime and mime.startswith("image/"):
            if self.file.__source_scheme__ == "url": 
                url = str(self.file)
            else:
                base64_data = base64.b64encode(self.file.read()).decode("utf-8")
                url = f"data:{mime};base64,{base64_data}"
            return {
                "type": "image_url", 
                "image_url": {"url": url},
            }

        # ! OpenAI API is broken. It errors with missing_required_parameter 'file_id'.
        elif mime == "application/pdf":
            # We need the base64 data of the pdf.
            data_url = File.serialize(self.file)
            base64_data = data_url.split(",", 1)[1]
            return {
                "type": "file",
                "file": {
                    "file_name": f"{uuid7()}.pdf",
                    "file_data": base64_data,
                },
            }

        elif mime and mime.startswith("audio/"):
            if self.file.__source_scheme__ == "data":
                base64_data = self.file.__source__.split(",", 1)[1]
            else:
                base64_data = base64.b64encode(self.file.read()).decode("utf-8")

            if "mp3" not in mime and "wav" not in mime:
                raise ValueError(f"Unsupported audio format: {mime}. Must be one of: mp3, wav")

            return {
                "type": "input_audio",
                "input_audio": {
                    "data": base64_data, 
                    "format": "wav" if "wav" in mime else "mp3",
                },
            }

        raise ValueError(f"Unsupported file type: {mime}. Must be an image or audio (wav/mp3) file.")


    def to_anthropic_input(self) -> dict[str, Any]:
        """Convert the file content to the input format required by Anthropic."""
        # Get mime type
        # source schemes: bytes, local_path, data, url
        if self.file.__source_scheme__ == "bytes":
            raise ValueError("Cannot convert bytes-source file to OpenAI message content.")
        elif self.file.__source_scheme__ == "data":
            mime = self.file.__source__.split(";")[0].split(":")[1]
        else:
            mime, _ = mimetypes.guess_type(str(self.file.__source__))

        if mime not in ["image/png", "image/jpeg", "image/webp", "image/gif", "application/pdf"]:
            raise ValueError(f"Unsupported file type: {mime}. Must be an image or pdf file.")

        data_url = File.serialize(self.file)
        base64_data = data_url.split(",", 1)[1]

        if mime.startswith("image/"):
            content_type = "image"
        else:
            content_type = "document"

        return {
            "type": content_type, 
            "source": {
                "type": "base64", 
                "media_type": mime, 
                "data": base64_data
            },
        }


class TextContent(Content):
    """
    This class represents a text content in a chat message.
    It also provides methods to convert the text content to the input format required by OpenAI and Anthropic.
    """

    type: Literal["text"] = "text"
    text: str 


    def to_openai_input(self) -> dict[str, Any]:
        """Convert the text content to the input format required by OpenAI."""
        return {
            "type": "text", 
            "text": self.text
        }


    def to_anthropic_input(self) -> dict[str, Any]:
        """Convert the text content to the input format required by Anthropic."""
        return {
            "type": "text", 
            "text": self.text
        }


class ToolUseContent(Content):
    """
    This class represents a tool use content in a chat message.
    It also provides methods to convert the tool use content to the input format required by OpenAI and Anthropic.
    """

    type: Literal["tool_use"] = "tool_use"
    id: str
    name: str
    input: dict[str, Any]


    def to_openai_input(self) -> dict[str, Any]:
        """Convert the tool use content to the input format required by OpenAI."""
        return {
            "id": self.id,
            "type": "function",
            "function": {
                "arguments": json.dumps(self.input),
                "name": self.name
            }
        }


    def to_anthropic_input(self) -> dict[str, Any]:
        """Convert the tool use content to the input format required by Anthropic."""
        return {
            "type": "tool_use",
            "id": self.id,
            "name": self.name,
            "input": self.input,
        }


class ToolResultContent(Content):
    """
    This class represents a tool result content in a chat message.
    It also provides methods to convert the tool result content to the input format required by OpenAI and Anthropic.
    """

    type: Literal["tool_result"] = "tool_result"
    id: str
    content: list[TextContent | FileContent]


    def to_openai_input(self) -> dict[str, Any]:
        """Convert the tool result content to the input format required by OpenAI."""
        return {
            "role": "tool",
            "content": [item.to_openai_input() for item in self.content],
            "tool_call_id": self.id
        }

    def to_anthropic_input(self) -> dict[str, Any]:
        """Convert the tool result content to the input format required by Anthropic."""
        return {
            "type": "tool_result",
            "tool_use_id": self.id,
            "content": [item.to_anthropic_input() for item in self.content],
        }



import inspect
from collections.abc import Callable
from typing import Any

import structlog
from pydantic import BaseModel, TypeAdapter

from ..state import RunContext
from ..types.models import create_model_from_argspec
from .base import BaseStep

logger = structlog.get_logger("timbal.graph.step")


class Step(BaseStep):
    """A class representing a processing step in a workflow.

    The Step class encapsulates a handler function and provides functionality to:
    - Validate input parameters using Pydantic models
    - Validate return values using Pydantic models
    - Execute the handler function with proper parameter passing
    - Support both synchronous and asynchronous execution

    The handler function's signature is automatically analyzed to create Pydantic models
    for parameter validation and return type checking. These models are accessible
    through the params_model() and return_model() methods.

    Attributes:
        handler_fn: The function that implements the step's processing logic
        handler_params_model: Pydantic model for validating input parameters
        handler_return_model: Pydantic model for validating return values
        is_llm: Flag indicating if this step involves LLM processing
        is_coroutine: Flag indicating if the handler is a coroutine function
        is_async_gen: Flag indicating if the handler is an async generator function
    """


    def __init__(
        self, 
        id: str,
        path: str | None = None,
        handler_fn: Callable[..., Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Overrides the __init__ method that is inherited from BaseStep, i.e. a pydantic BaseModel.

        Args:
            id: The unique identifier for the step (within a flow or subflow).
            path: Optional path for the flow. Use when directly adding this to a flow or subflow.
            handler_fn: The handler function to be used.
            **kwargs: Additional keyword arguments will be passed to the BaseStep's __init__ method.
        """
        if path is None:
            path = id

        super().__init__(id=id, path=path, **kwargs)
        
        if not inspect.isfunction(handler_fn):
            raise TypeError(f"handler_fn must be a function, got {type(handler_fn)}")

        self.handler_fn = handler_fn
        handler_argspec = inspect.getfullargspec(handler_fn)
        self.handler_params_model = create_model_from_argspec(
            name=f"Step_{self.id}_params", 
            argspec=handler_argspec
        )
        self.handler_params_model_schema = self.handler_params_model.model_json_schema()

        self.handler_return_model = handler_argspec.annotations.get("return", Any)
        self.handler_return_model_schema = TypeAdapter(self.handler_return_model).json_schema()

        # Initialize as False, it will be set to True in the Flow.add_llm method.
        self.is_llm = False
        # These are used to determine if we need to run the step in an executor.
        self.is_coroutine = inspect.iscoroutinefunction(handler_fn)
        self.is_async_gen = inspect.isasyncgenfunction(handler_fn)


    def params_model(self) -> BaseModel:
        """Returns the Pydantic model defining the expected parameters for this step."""
        return self.handler_params_model
        

    def params_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the step's parameter model."""
        return self.handler_params_model_schema
    

    def return_model(self) -> Any:
        """Returns the expected return type for this step."""
        return self.handler_return_model


    def return_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the step's return value model."""
        return self.handler_return_model_schema


    def run(
        self, 
        context: RunContext | None = None, # noqa: ARG002
        **kwargs: Any,
    ) -> Any:
        """Executes the step's processing logic."""
        return self.handler_fn(**kwargs)
    